"Internal support for the rdkit toolkit"

# Copyright (c) 2010-2020 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See http://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

from __future__ import absolute_import

import sys

from ._compat import BytesIO, unicode_type, bytes_type, next, enumerate

import rdkit
import rdkit.rdBase
from rdkit import Chem

from . import _compat

from . import io
from . import base_toolkit
from . import _text_toolkit

# If the attribute doesn't exist then this is an unsupported pre-2010 RDKit distribution
SOFTWARE = "RDKit/" + getattr(rdkit.rdBase, "rdkitVersion", "unknown")

EXTERNAL_NAME = "RDKit"

# Check for compatibility with older versions of RDKit

# older RDKit versions don't handle Unicode
mol = Chem.MolFromSmiles("C")
try:
    mol.SetProp(u"A", u"B")
except Exception:
    is_unicode_aware = False
    def _SetProp(mol, key, value):
        mol.SetProp(key.encode("utf8"), value.encode("utf8"))
    def _GetProp(mol, key):
        return mol.GetProp(key.encode("utf8")).decode("utf8")
else:
    is_unicode_aware = True
    def _SetProp(mol, key, value):
        mol.SetProp(key, value)
    def _GetProp(mol, key):
        return mol.GetProp(key)
del mol


MolFromMolBlock = Chem.MolFromMolBlock

# If InChI isn't installed then make a fake exception
if hasattr(Chem, "MolFromInchi"):
    has_inchi = True
    from rdkit.Chem.inchi import InchiReadWriteError

    # Support older versions of RDKit which don't support Unicode
    try:
        Chem.MolFromInchi(u"InChI=1/U/")
    except Exception:
        def MolFromInchi(inchi, **inchi_kwargs):
            if isinstance(inchi, unicode_type):
                inchi = inchi.encode("utf8")
            return Chem.MolFromInchi(inchi, **inchi_kwargs)
    else:
        MolFromInchi = Chem.MolFromInchi
    
else:
    has_inchi = False
    class InchiReadWriteError(Exception):
        pass
        
# Support older versions of RDKit which don't support Unicode
try:
    Chem.MolFromSmiles(u"C")
except Exception:
    def MolFromSmiles(smiles, sanitize):
        if isinstance(smiles, unicode_type):
            smiles = smiles.encode("utf8")
        return Chem.MolFromSmiles(smiles, sanitize)
else:
    MolFromSmiles = Chem.MolFromSmiles



# The "allBondsExplicit" option was added to version control on 2012-03-26.
# This workaround lets me support older toolkits.
_mol = Chem.MolFromSmiles("C")
try:
    Chem.MolToSmiles(_mol, allBondsExplicit=True)
except Exception:
    if isinstance(Chem.MolToSmiles(_mol), type(u"")):
        def MolToSmiles(mol, isomericSmiles=False, kekuleSmiles=False, canonical=True,
                            allBondsExplicit=False, allHsExplicit=False):
            return Chem.MolToSmiles(mol, isomericSmiles = isomericSmiles,
                                    kekuleSmiles = kekuleSmiles, canonical = canonical)
    else:
        def MolToSmiles(mol, isomericSmiles=False, kekuleSmiles=False, canonical=True,
                            allBondsExplicit=False, allHsExplicit=False):
            return Chem.MolToSmiles(mol, isomericSmiles = isomericSmiles,
                                    kekuleSmiles = kekuleSmiles, canonical = canonical).decode("utf8")
        
else:
    # allHsExplicit is not in my test build from 2016.09.3
    # XXX When was it added?
    try:
        Chem.MolToSmiles(_mol, allBondsExplicit=True, allHsExplicit=True)
    except Exception:
        if isinstance(Chem.MolToSmiles(_mol), type(u"")):
            def MolToSmiles(mol, isomericSmiles=False, kekuleSmiles=False, canonical=True,
                                allBondsExplicit=False, allHsExplicit=False):
                return Chem.MolToSmiles(mol, isomericSmiles = isomericSmiles,
                                        kekuleSmiles = kekuleSmiles, canonical = canonical,
                                        allBondsExplicit = allBondsExplicit)
        else:
            def MolToSmiles(mol, isomericSmiles=False, kekuleSmiles=False, canonical=True,
                                allBondsExplicit=False, allHsExplicit=False):
                return Chem.MolToSmiles(mol, isomericSmiles = isomericSmiles,
                                        kekuleSmiles = kekuleSmiles, canonical = canonical,
                                        allBondsExplicit = allBondsExplicit).decode("utf8")
    else:
        if isinstance(Chem.MolToSmiles(_mol), type(u"")):
            MolToSmiles = Chem.MolToSmiles
        else:
            def MolToSmiles(mol, isomericSmiles=False, kekuleSmiles=False, canonical=True,
                                allBondsExplicit=False, allHsExplicit=False):
                return Chem.MolToSmiles(mol, isomericSmiles = isomericSmiles, kekuleSmiles = kekuleSmiles,
                                        canonical = canonical, allBondsExplicit = allBondsExplicit,
                                        allHsExplicit = allHsExplicit).decode("utf8")

MolToCXSmiles = getattr(Chem, "MolToCXSmiles", None)
if MolToCXSmiles is None:
    def MolToCXSmiles(mol, **kwargs):
        global MolToCXSmiles
        sys.stderr.write("This version of RDKit does not have Chem.MolToCXSmiles(), using Chem.MolToSmiles()\n")
        MolToCXSmiles = MolToSmiles
        return MolToSmiles(mol, **kwargs)

MolToV3KMolBlock = getattr(Chem, "MolToV3KMolBlock", None)
if MolToV3KMolBlock is None:
    def MolToV3KMolBlock(mol, **kwargs):
        global MolToV3KMolBlock
        import warnings
        warnings.warn("This version of RDKit does not have Chem.MolToV3KMolBlock(), using Chem.MolToMolBlock()")
        MolToV3KMolBlock = Chem.MolToMolBlock
        return Chem.MolToMolBlock(mol, **kwargs)

# Older versions of RDKit's Mol2 reader don't support cleanupSubstructures
if "cleanupSubstructures" in Chem.MolFromMol2Block.__doc__:
    MolFromMol2Block = Chem.MolFromMol2Block
else:
    _MolFromMol2Block = Chem.MolFromMol2Block
    def MolFromMol2Block(record, sanitize, removeHs, cleanupSubstructures):
        return _MolFromMol2Block(record, sanitize, removeHs)
    
# Older versions of RDKit don't support
MolSanitizeException = getattr(Chem, "MolSanitizeException", None)
if MolSanitizeException is None:
    # The code will use a fall-back path.
    class MolSanitizeException(Exception):
        pass

def MolFromSDBlock(molBlock, sanitize=True, removeHs=True, strictParsing=True, includeTags=True):
    mol = MolFromMolBlock(molBlock, sanitize, removeHs, strictParsing)
    if mol is None:
        return mol

    # MolFromMolBlock reader doesn't process the properties.
    # According to
    #   http://www.mail-archive.com/rdkit-discuss@lists.sourceforge.net/msg01436.html
    # I can make a new SDMolSupplier, then SetData(), get the first
    # record, and get its property names. Not going to do that.
    # Instead, read the tag data myself.
    if includeTags:
        if isinstance(molBlock, bytes_type):
            for k, v in _text_toolkit.get_sdf_tag_pairs(molBlock):
                if k[:1] != b"_":
                    try:
                        k = k.decode("utf8")
                    except UnicodeDecodeError as err:
                        err.reason = err.reason + " of tag name %r" % (k,)
                        raise
                    try:
                        v = v.decode("utf8")
                    except UnicodeDecodeError as err:
                        err.reason = err.reason + " of tag %r value %r" % (k, v)
                        raise
                    _SetProp(mol, k, v)
        else:
            for k, v in _text_toolkit.get_sdf_tag_pairs(molBlock):
                if k[:1] != "_":
                    _SetProp(mol, k, v)

    return mol

def _or_string(location):
    filename = location.filename
    if filename is None:
        return "<string>"
    return filename

def _or_bytes(location):
    filename = location.filename
    if filename is None:
        return "<bytes>"
    return filename



class RDKitFormatConfig(object):
    supports_io = True
    alternate_extensions = ()
    is_available = True

    def create_bytes(self, mol, id, writer_args, error_handler):
        s = self.create_string(mol, id, writer_args, error_handler)
        if s is not None:
            return s.encode("utf8")
        return None
    
def iter_read_delimited_fields_as_strings(infile, close, has_header, delimiter, error_handler, location, format_name,
                                         encoding, encoding_errors):
    it = iter_read_delimited_fields_as_bytes(infile, close, has_header, delimiter, error_handler, location, format_name)
    
    def iter_read_delimited_fields_as_strings(it, error_handler, format_name, encoding, encoding_errors):
        for field1_bytes, field2_bytes in it:
            if field1_bytes is None:
                field1 = None
            else:
                try:
                    field1 = field1_bytes.decode(encoding, encoding_errors)
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the first field from the %s file, %r: %s"
                                        % (format_name, field1_bytes, err), location)
                    continue
            if field2_bytes is None:
                field2 = None
            else:
                try:
                    field2 = field2_bytes.decode(encoding, encoding_errors)
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the second field from the %s file, %r: %s"
                                        % (format_name, field2_bytes, err), location)
                    continue
            yield field1, field2
                
    return iter_read_delimited_fields_as_strings(it, error_handler, format_name, encoding, encoding_errors)

def iter_read_delimited_fields_as_bytes(infile, close, has_header, delimiter, error_handler, location, format_name):
    it = _iter_read_delimited_fields(infile, close, has_header, delimiter, error_handler, location, format_name)
    next(it)
    return it

def _iter_read_delimited_fields(infile, close, has_header, delimiter, error_handler, location, format_name):
    lineno = start_offset = end_offset = 0
    record = None
    def get_lineno():
        return lineno
    def get_recno():
        return lineno
    def get_offsets():
        return start_offset, end_offset
    def get_record():
        return record
    
    location.register(get_lineno = get_lineno,
                      get_recno = get_recno,
                      get_offsets = get_offsets,
                      get_record = get_record,
        )

    start_offset = 0
    split_two_terms_bytes = delimiter.split_two_terms_bytes
    try:
        yield  "Ready for processing!"
        file_iter = enumerate(infile, 1)
        if has_header:
            # Skip the first line
            header_line = next(file_iter, None)
            if header_line is None:
                return
            
        for lineno, record in file_iter:
            start_offset = end_offset
            end_offset = start_offset + len(record)
            line = record.rstrip(b"\r\n")
            terms = split_two_terms_bytes(line)

            if not terms:
                error_handler.error("Empty line in %s file" % (format_name,), location)
                continue

            yield terms
    finally:
        if close is not None:
            close()
        location.save(lineno = get_lineno(),
                      recno = get_recno(),
                      offsets = None,
                      record = None,
            )

    
def iter_read_smiles_ids_and_molecules(infile, close, sanitize, has_header, delimiter, error_handler, location,
                                       encoding, encoding_errors):
    it = _iter_read_smiles_ids_and_molecules(infile, close, sanitize, has_header, delimiter, error_handler, location,
                                             encoding, encoding_errors)
    next(it)
    return it
    
def _iter_read_smiles_ids_and_molecules(infile, close, sanitize, has_header, delimiter, error_handler, location,
                                        encoding, encoding_errors):
    mol = None
    def get_mol():
        return mol
    location.register(get_mol=get_mol)
    iter_reader = iter_read_delimited_fields_as_strings(infile, close, has_header, delimiter, error_handler, location,
                                                        "SMILES", encoding, encoding_errors)
    try:
        yield  "Ready for processing!"
        for smiles, id in iter_reader:
            mol = MolFromSmiles(smiles, sanitize)
            if mol is None:
                error_handler.error("RDKit cannot parse the SMILES %s" % (_compat.myrepr(smiles),), location)
                continue
            else:
                if id is not None:
                    _SetProp(mol, u"_Name", id)
            yield id, mol
            mol = None  # reset in case the lower-level handler skips a record/raises an error
    finally:
        # Don't save the final molecule.
        location.save(mol=None)
        mol = None
        iter_reader.close()
        
        
class BaseSmiFormat(RDKitFormatConfig):
    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler, location,
                               encoding, encoding_errors):
        delimiter = base_toolkit.get_delimiter(self, reader_args, "whitespace")
        sanitize = bool(reader_args["sanitize"])
        has_header = bool(reader_args["has_header"])

        infile, close = io.open_compressed_input(source, compression, EXTERNAL_NAME)

        iter_reader = iter_read_smiles_ids_and_molecules(infile, close, sanitize, has_header, delimiter,
                                                         error_handler, location, encoding, encoding_errors)
        metadata = base_toolkit.FormatMetadata(location.filename, self.record_format,
                                               {"delimiter": delimiter.name,
                                                "sanitize": sanitize,
                                                "has_header": has_header})
        
        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)


    def read_ids_and_molecules_from_string(self, content, compression,
                                           id_tag, reader_args, error_handler, location):
        assert not compression, compression
        delimiter = base_toolkit.get_delimiter(self, reader_args, "whitespace")
        sanitize = bool(reader_args["sanitize"])
        has_header = bool(reader_args["has_header"])

        infile = BytesIO(content)
        close = infile.close

        iter_reader = iter_read_smiles_ids_and_molecules(infile, close, sanitize, has_header, delimiter,
                                                         error_handler, location, "utf8", "strict")
        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format,
                                               {"delimiter": delimiter.name,
                                                "sanitize": sanitize,
                                                "has_header": has_header})

        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)


    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, reader_args, "whitespace")
        split_two_terms_bytes = delimiter.split_two_terms_bytes
        split_two_terms_string = delimiter.split_two_terms_string
        sanitize = bool(reader_args["sanitize"])

        def rdkit_smiles_parser(content):
            if isinstance(content, unicode_type):
                first_line, _, _ = content.partition("\n") # get the first record from the input
                terms = split_two_terms_string(first_line.rstrip("\r"))
            
                if terms is None:
                    error_handler.error("The empty string is not a valid SMILES input")
                    return None, None

                smiles, id = terms
                mol = MolFromSmiles(smiles, sanitize)
                if mol is None:
                    error_handler.error("RDKit cannot parse the SMILES %s" % (_compat.myrepr(smiles),))
                else:
                    if id is not None:
                        _SetProp(mol, u"_Name", id)
                return id, mol
            else:
                first_line, _, _ = content.partition(b"\n") # get the first record from the input
                terms = split_two_terms_bytes(first_line.rstrip(b"\r"))
            
                if terms is None:
                    error_handler.error("The empty string is not a valid SMILES input")
                    return None, None

                smiles_bytes, id_bytes = terms
                mol = MolFromSmiles(smiles_bytes, sanitize)
                if mol is None:
                    error_handler.error("RDKit cannot parse the SMILES %s" % (_compat.myrepr(smiles_bytes),))
                    return None, None
                else:
                    if id_bytes is not None:
                        try:
                            id = id_bytes.decode("utf8")
                        except UnicodeDecodeError as err:
                            error_handler.error("Cannot decode the identifier %r: %s"
                                                % (id_bytes, err))
                            return None, None
                        _SetProp(mol, u"_Name", id)
                    else:
                        id = None
                return id, mol

        rdkit_smiles_parser.__name__ = "rdkit_" + self.register_name + "_parse_id_and_molecule"
        return rdkit_smiles_parser
        

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, reader_args, "whitespace")

        if isinstance(content, unicode_type):
            first_line, _, _ = content.partition("\n") # get the first record from the input
            terms = delimiter.split_two_terms_string(first_line.rstrip("\r"))

            if terms is None:
                error_handler.error("The empty string is not a valid SMILES input")
                return None, None

            smiles, id = terms
            mol = MolFromSmiles(smiles, reader_args["sanitize"])
            if mol is None:
                error_handler.error("RDKit cannot parse the SMILES %s" % (_compat.myrepr(smiles),))
                return None, None
            
            if id is not None:
                _SetProp(mol, u"_Name", id)
            return id, mol
        else:
            first_line, _, _ = content.partition(b"\n") # get the first record from the input
            terms = delimiter.split_two_terms_bytes(first_line.rstrip(b"\r"))

            if terms is None:
                error_handler.error("The empty string is not a valid SMILES input")
                return None, None

            smiles_bytes, id_bytes = terms
            mol = MolFromSmiles(smiles_bytes, reader_args["sanitize"])
            if mol is None:
                error_handler.error("RDKit cannot parse the SMILES %s" % (_compat.myrepr(smiles_bytes),))
                return None, None
            
            if id_bytes is not None:
                try:
                    id = id_bytes.decode("utf8")
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the identifier %r: %s"
                                        % (id_bytes, err))
                    return None, None
                _SetProp(mol, u"_Name", id)
            else:
                id = None
            return id, mol
            


    def open_molecule_writer(self, destination, outfile, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        delimiter = base_toolkit.get_delimiter(self, writer_args, "whitespace")
        to_smiles_kwargs = {
            "isomericSmiles": bool(writer_args["isomericSmiles"]),
            "kekuleSmiles": bool(writer_args["kekuleSmiles"]),
            "canonical": bool(writer_args["canonical"]),
            "allBondsExplicit": bool(writer_args["allBondsExplicit"]),
            "allHsExplicit": bool(writer_args["allHsExplicit"]),
            "cxsmiles": bool(writer_args["cxsmiles"]),
            }
        new_writer_args = to_smiles_kwargs.copy()
        new_writer_args["delimiter"] = delimiter.name
        
        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_writer_args)
        
        return base_toolkit.MoleculeWriter(
            metadata,
            gen_write_smi_structures(outfile, close, error_handler, location, delimiter, to_smiles_kwargs),
            location)
            

    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        delimiter = base_toolkit.get_delimiter(self, writer_args, "whitespace")
        to_smiles_kwargs = {
            "isomericSmiles": bool(writer_args["isomericSmiles"]),
            "kekuleSmiles": bool(writer_args["kekuleSmiles"]),
            "canonical": bool(writer_args["canonical"]),
            "allBondsExplicit": bool(writer_args["allBondsExplicit"]),
            "allHsExplicit": bool(writer_args["allHsExplicit"]),
            "cxsmiles": bool(writer_args["cxsmiles"]),
            }
        new_writer_args = to_smiles_kwargs.copy()
        new_writer_args["delimiter"] = delimiter.name
        outfile, close, getvalue = io._open_output_string()

        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_smi_structures(outfile, close, error_handler, location, delimiter, to_smiles_kwargs),
            getvalue, location)
            

    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        delimiter = base_toolkit.get_delimiter(self, writer_args, "whitespace")
        to_smiles_kwargs = {
            "isomericSmiles": bool(writer_args["isomericSmiles"]),
            "kekuleSmiles": bool(writer_args["kekuleSmiles"]),
            "canonical": bool(writer_args["canonical"]),
            "allBondsExplicit": bool(writer_args["allBondsExplicit"]),
            "allHsExplicit": bool(writer_args["allHsExplicit"]),
            "cxsmiles": bool(writer_args["cxsmiles"]),
            }
        new_writer_args = to_smiles_kwargs.copy()
        new_writer_args["delimiter"] = delimiter.name
        outfile, close, getvalue = io._open_compressed_output_bytes(compression, EXTERNAL_NAME, level)
    
        metadata = base_toolkit.FormatMetadata(_or_bytes(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_smi_structures(outfile, close, error_handler, location, delimiter, to_smiles_kwargs),
            getvalue,
            location)
            

    def create_string(self, mol, id, writer_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, writer_args, "whitespace")
        if writer_args["cxsmiles"]:
            ToSmiles = MolToCXSmiles
        else:
            ToSmiles = MolToSmiles
        smiles = ToSmiles(mol,
                          isomericSmiles = writer_args["isomericSmiles"],
                          kekuleSmiles = writer_args["kekuleSmiles"],
                          canonical = writer_args["canonical"],
                          allBondsExplicit = writer_args["allBondsExplicit"],
                          allHsExplicit = writer_args["allHsExplicit"],
                          )

        if id is None:
            try:
                id = _GetProp(mol, u"_Name")
            except KeyError:
                id = ""

        return delimiter.get_format_string() % (smiles, id)
    

_smi_input_arg_info = (
    ("sanitize", base_toolkit.parse_boolean),
    )

_smi_input_defaults = {
    "sanitize": True,
    }

_smi_output_arg_info = (
    ("isomericSmiles", base_toolkit.parse_boolean),
    ("kekuleSmiles", base_toolkit.parse_boolean),
    ("canonical", base_toolkit.parse_boolean),
    ("allBondsExplicit", base_toolkit.parse_boolean),
    ("allHsExplicit", base_toolkit.parse_boolean),
    ("cxsmiles", base_toolkit.parse_boolean),
    )
_smi_output_defaults = {
    "isomericSmiles": False,
    "kekuleSmiles": False,
    "canonical": False,
    "allBondsExplicit": False,
    "allHsExplicit": False,
    "cxsmiles": False,
    }


# The three categories follow OEChem 2.0's lead:
#   "smi" = isomeric, canonical (alternate is "ism")
#   "can" = non-isomeric, canonical
#   "usm" = non-isomeric, non-canonical

_parse_delimiter = base_toolkit.DelimiterParser("whitespace")

class SmiFormat(BaseSmiFormat):
    register_name = "smi"
    alternate_extensions = ("ism", "isosmi")
    is_available = True
    record_format = "smi"

    class input(object):
        arg_info = _smi_input_arg_info + (
            ("has_header", base_toolkit.parse_boolean),
            ("delimiter", _parse_delimiter),
            )
        defaults = _smi_input_defaults.copy()
        defaults.update({
            "has_header": False,
            "delimiter": None,
            })
    
    class output(object):
        arg_info = _smi_output_arg_info + (
            ("delimiter", _parse_delimiter),
            )
        defaults = _smi_output_defaults.copy()
        defaults.update({
            "isomericSmiles": True,   # Note: I do not follow the RDKit default of False
            "canonical": True,         # RDKit's default is canonical
            "delimiter": None,
            })

class CanFormat(BaseSmiFormat):
    register_name = record_format = "can"
    is_available = True

    input = SmiFormat.input
    
    class output(object):
        arg_info = SmiFormat.output.arg_info
        defaults = SmiFormat.output.defaults.copy()
        defaults.update({
            "isomericSmiles": False,
            "canonical": True,         # This is the canonical SMILES
            })

class UsmFormat(BaseSmiFormat):
    register_name = record_format = "usm"
    is_available = True

    input = SmiFormat.input
    
    class output(object):
        arg_info = SmiFormat.output.arg_info
        defaults = SmiFormat.output.defaults.copy()
        defaults.update({
            "isomericSmiles": False,    # not isomeric ...
            "canonical": False,         #   ... and not canonical
            })


def gen_write_smi_structures(outfile, close, error_handler, location, delimiter, to_kwargs):
    if to_kwargs.pop("cxsmiles", False):
        ToSmiles = MolToCXSmiles
    else:
        ToSmiles = MolToSmiles
    gen = _gen_write_delimited_structures(outfile, close, error_handler, location, delimiter,
                                          True, ToSmiles, "SMILES", to_kwargs)
    next(gen)
    return gen

def gen_write_inchi_structures(outfile, close, error_handler, location, delimiter, include_id, to_kwargs):
    gen = _gen_write_delimited_structures(outfile, close, error_handler, location, delimiter,
                                          include_id, Chem.MolToInchi, "InChI", to_kwargs)
    next(gen)
    return gen

def _MolToInchiKey(mol, options, logLevel, treatWarningAsError):
    inchi = Chem.MolToInchi(mol, options=options, logLevel=logLevel, treatWarningAsError=treatWarningAsError)
    return Chem.InchiToInchiKey(inchi)

def gen_write_inchikey_structures(outfile, close, error_handler, location, delimiter, include_id, to_kwargs):
    gen = _gen_write_delimited_structures(outfile, close, error_handler, location, delimiter,
                                          include_id, _MolToInchiKey, "InChIKey", to_kwargs)
    next(gen)
    return gen

def _gen_write_delimited_structures(outfile, close, error_handler, location, delimiter,
                                    include_id, mol_to_string, format_label, to_kwargs):
    recno = output_recno = 0
    def get_recno():
        return recno
    def get_output_recno():
        return output_recno

    location.register(get_recno=get_recno,
                      get_output_recno=get_output_recno)

    fmt = delimiter.get_format_string()
    
    try:
        yield_value = None
        while 1:
            command, args = yield yield_value
            yield_value = None
            try:
                if command == "write_molecules":
                    for mol in args:
                        recno += 1
                        try:
                            s = mol_to_string(mol, **to_kwargs)
                        except InchiReadWriteError as err:
                            # This can happen if InChI generates a warning and
                            # "treatWarningAsError" is True.
                            error_handler.error("RDKit cannot create the %s string: %s"
                                                % (format_label, err.args[1]), location)
                            continue
                        
                        except MolSanitizeException as err:
                            error_handler.error("RDKit cannot create the %s string: %s"
                                                    % (format_label, err), location)
                            continue
                        
                        except ValueError as err:
                            # It looks like the molecule is sanitized before passing
                            # the molecule to InChI, and sanitization can raise a
                            # ValueError. But so can passing in the wrong value.
                            msg = str(err)
                            if msg.startswith("Sanitization error"):
                                error_handler.error("RDKit cannot create the %s string: %s"
                                                    % (format_label, msg), location)
                                continue
                            else:
                                raise
                        if not s:
                            error_handler.error("RDKit cannot create the %s string" % (format_label,),
                                                location)
                            continue
                        if include_id:
                            try:
                                id = _GetProp(mol, u"_Name")
                            except KeyError:
                                id = ""
                                    
                            if not outfile.write_string(fmt % (s, id), error_handler, location):
                                continue
                        else:
                            if not outfile.write_string(s + "\n", error_handler, location):
                                continue
                        output_recno += 1

                elif command == "write_ids_and_molecules":
                    for id, mol in args:
                        recno += 1
                        try:
                            s = mol_to_string(mol, **to_kwargs)
                        except InchiReadWriteError as err:
                            # This can happen if InChI generates a warning and
                            # "treatWarningAsError" is True.
                            error_handler.error("RDKit cannot create the %s string: %s"
                                                % (format_label, err.args[1]), location)
                            continue
                        
                        except MolSanitizeException as err:
                            error_handler.error("RDKit cannot create the %s string: %s"
                                                % (format_label, err), location)
                            continue
                        
                        except ValueError as err:
                            # It looks like the molecule is sanitized before passing
                            # the molecule to InChI, and sanitization can raise a
                            # ValueError. But so can passing in the wrong value.
                            msg = str(err)
                            if msg.startswith("Sanitization error"):
                                error_handler.error("RDKit cannot create the %s string: %s"
                                                    % (format_label, msg), location)
                                continue
                            else:
                                raise
                        if not s:
                            error_handler.error("RDKit cannot create the %s string" % (format_label,),
                                                location)
                            continue
                        if include_id:
                            if id is None:
                                try:
                                    id = _GetProp(mol, u"_Name")
                                except KeyError:
                                    id = ""
                            if not outfile.write_string(fmt % (s, id), error_handler, location):
                                continue
                        else:
                            if not outfile.write_string(s + "\n", error_handler, location):
                                continue
                        output_recno += 1

                elif command == "write_molecule":
                    id, mol = args
                    recno += 1
                    try:
                        s = mol_to_string(mol, **to_kwargs)
                    except InchiReadWriteError as err:
                        # This can happen if InChI generates a warning and
                        # "treatWarningAsError" is True.
                        error_handler.error("RDKit cannot create the %s string: %s"
                                            % (format_label, err.args[1]), location)
                        continue
                    except MolSanitizeException as err:
                        error_handler.error("RDKit cannot create the %s string: %s"
                                            % (format_label, err), location)
                        continue

                    except ValueError as err:
                        # It looks like the molecule is sanitized before passing
                        # the molecule to InChI, and sanitization can raise a
                        # ValueError. But so can passing in the wrong value.
                        msg = str(err)
                        if msg.startswith("Sanitization error"):
                            error_handler.error("RDKit cannot create the %s string: %s"
                                                % (format_label, msg), location)
                            continue
                        else:
                            raise
                    if not s:
                        error_handler.error("RDKit cannot create the %s string" % (format_label,),
                                            location)
                        continue
                    if include_id:
                        if id is None:
                            try:
                                id = _GetProp(mol, u"_Name")
                            except KeyError:
                                id = ""

                        if not outfile.write_string(fmt % (s, id), error_handler, location):
                            continue
                    else:
                        if not outfile.write_string(s + "\n", error_handler, location):
                            continue
                    output_recno += 1

                else:
                    raise AssertionError("Unknown command: %r" % (command,))
            except Exception as err:
                yield_value = err, sys.exc_info()[2]
                        
    finally:
        if close is not None:
            close()
        location.save(output_recno = get_output_recno(),
                      recno = get_recno())
            

def iter_read_inchi_structures(infile, close, has_header, delimiter, inchi_kwargs, error_handler, location,
                               encoding, encoding_errors):
    x = _iter_read_inchi_structures(infile, close, has_header, delimiter, inchi_kwargs, error_handler, location,
                                    encoding, encoding_errors)
    next(x)
    return x

def _iter_read_inchi_structures(infile, close, has_header, delimiter, inchi_kwargs, error_handler, location,
                                encoding, encoding_errors):
    mol = None
    def get_mol():
        return mol
    location.register(get_mol=get_mol)
    delimited_field_reader = iter_read_delimited_fields_as_strings(infile, close, has_header, delimiter, error_handler,
                                                                   location, "InChI", encoding, encoding_errors)
    try:
        yield "Ready for processing!"
        for inchi, id in delimited_field_reader:
            mol = MolFromInchi(inchi, **inchi_kwargs)
            if mol is None:
                error_handler.error("RDKit cannot parse the InChI %r" % (inchi,), location)
                continue
            else:
                if id is not None:
                    _SetProp(mol, u"_Name", id)
            yield id, mol
            mol = None # reset in case the lower-level handler skips a record/raises an error
    finally:
        location.save(mol=None)
        mol = None
        delimited_field_reader.close()
    

# Format is:
#    <inchi> <delimiter> <id>

_inchi_input_arg_info = (
    ("sanitize", base_toolkit.parse_boolean),
    ("removeHs", base_toolkit.parse_boolean),
    ("logLevel", int),
    ("treatWarningAsError", base_toolkit.parse_boolean),
    )
_inchi_input_defaults = {
    "sanitize": True,
    "removeHs": True,
    "logLevel": None,
    "treatWarningAsError": False,
}
_inchi_output_arg_info = (
    ("options", str),  # Any way to standardize this with OEChem?
    ("logLevel", int),
    ("treatWarningAsError", base_toolkit.parse_boolean),
    )
_inchi_output_defaults = {
    "options": "",
    "logLevel": None,
    "treatWarningAsError": False,
    }
    

class InChIFormat(RDKitFormatConfig):
    register_name = record_format = "inchi"
    is_available = has_inchi

    class input(object):
        arg_info = _inchi_input_arg_info + (
            ("delimiter", _parse_delimiter),
            )
        defaults = _inchi_input_defaults.copy()
        defaults["delimiter"] = None
        
    class output(object):
        arg_info = _inchi_output_arg_info + (
            ("delimiter", _parse_delimiter),
            ("include_id", base_toolkit.parse_boolean),
            )
        defaults = _inchi_output_defaults.copy()
        defaults["delimiter"] = None
        defaults["include_id"] = True

    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler, location,
                               encoding, encoding_errors):
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        inchi_kwargs = {
            "sanitize": bool(reader_args["sanitize"]),
            "removeHs": bool(reader_args["removeHs"]),
            "logLevel": reader_args["logLevel"],
            "treatWarningAsError": bool(reader_args["treatWarningAsError"]),
            }
        new_reader_args = inchi_kwargs.copy()
        new_reader_args["delimiter"] = delimiter.name

        infile, close = io.open_compressed_input(source, compression, EXTERNAL_NAME)

        iter_reader = iter_read_inchi_structures(infile, close, False, delimiter, inchi_kwargs,
                                                 error_handler, location, encoding, encoding_errors)

        metadata = base_toolkit.FormatMetadata(location.filename, self.record_format, new_reader_args)
        
        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)

    def read_ids_and_molecules_from_string(self, content, compression,
                                            id_tag, reader_args, error_handler, location):
        assert not compression, compression
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        inchi_kwargs = {
            "sanitize": bool(reader_args["sanitize"]),
            "removeHs": bool(reader_args["removeHs"]),
            "logLevel": reader_args["logLevel"],
            "treatWarningAsError": bool(reader_args["treatWarningAsError"]),
            }
        new_reader_args = inchi_kwargs.copy()
        new_reader_args["delimiter"] = delimiter.name

        infile = BytesIO(content)
        close = infile.close

        iter_reader = iter_read_inchi_structures(infile, close, False, delimiter, inchi_kwargs,
                                                error_handler, location, "utf8", "strict")
        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format, new_reader_args)
        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)


    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        split_two_terms_bytes = delimiter.split_two_terms_bytes
        split_two_terms_string = delimiter.split_two_terms_string
        inchi_kwargs = {
            "sanitize": bool(reader_args["sanitize"]),
            "removeHs": bool(reader_args["removeHs"]),
            "logLevel": reader_args["logLevel"],
            "treatWarningAsError": bool(reader_args["treatWarningAsError"]),
            }

        def rdkit_inchi_parse_id_and_molecule(content):
            if isinstance(content, unicode_type):
                first_line, _, _ = content.partition("\n") # get the first record from the input
                terms = split_two_terms_string(first_line.rstrip("\r"))
            
                if terms is None:
                    error_handler.error("The empty string is not a valid InChI input")
                    return None, None

                inchi, id = terms
                mol = MolFromInchi(inchi, **inchi_kwargs)
                if mol is None:
                    error_handler.error("RDKit cannot parse the InChI %r" % (inchi,))
                    return None, None
                
                if id is not None:
                    _SetProp(mol, u"_Name", id)
                return id, mol
            else:
                first_line, _, _ = content.partition(b"\n") # get the first record from the input
                terms = split_two_terms_bytes(first_line.rstrip(b"\r"))
            
                if terms is None:
                    error_handler.error("The empty string is not a valid InChI input")
                    return None, None

                inchi_bytes, id_bytes = terms
                mol = MolFromInchi(inchi_bytes, **inchi_kwargs)
                if mol is None:
                    error_handler.error("RDKit cannot parse the InChI %r" % (inchi_bytes,))
                    return None, None
                else:
                    if id_bytes is None:
                        id = None
                    else:
                        try:
                            id = id_bytes.decode("utf8")
                        except UnicodeDecodeError as err:
                            error_handler.error("Cannot decode the identifier %r: %s"
                                                % (id_bytes, err))
                            return None, None
                        _SetProp(mol, u"_Name", id)
                return id, mol
            
        return rdkit_inchi_parse_id_and_molecule
        

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        inchi_kwargs = {
            "sanitize": bool(reader_args["sanitize"]),
            "removeHs": bool(reader_args["removeHs"]),
            "logLevel": reader_args["logLevel"],
            "treatWarningAsError": reader_args["treatWarningAsError"],
            }

        if isinstance(content, unicode_type):
            first_line, _, _ = content.partition("\n") # get the first record from the input
            terms = delimiter.split_two_terms_string(first_line.rstrip("\r"))
            
            if terms is None:
                error_handler.error("The empty string is not a valid InChI input")
                return None, None
            
            inchi, id = terms
            mol = MolFromInchi(inchi, **inchi_kwargs)
            if mol is None:
                error_handler.error("RDKit cannot parse the InChI %r" % (inchi,))
                return None, None
            if id is not None:
                _SetProp(mol, u"_Name", id)
            return id, mol
        else:
            first_line, _, _ = content.partition(b"\n") # get the first record from the input
            terms = delimiter.split_two_terms_bytes(first_line.rstrip(b"\r"))
            
            if terms is None:
                error_handler.error("The empty string is not a valid InChI input")
                return None, None
            
            inchi_bytes, id_bytes = terms
            mol = MolFromInchi(inchi_bytes, **inchi_kwargs)
            if mol is None:
                error_handler.error("RDKit cannot parse the InChI %r" % (inchi_bytes,))
                return None, None

            if id_bytes is not None:
                try:
                    id = id_bytes.decode("utf8")
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the identifier %r: %s"
                                        % (id_bytes, err))
                    return None, None
                _SetProp(mol, u"_Name", id)
            return id, mol
    
    def open_molecule_writer(self, destination, outfile, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        inchi_kwargs = {
            "options": writer_args["options"],
            "logLevel": writer_args["logLevel"],
            "treatWarningAsError": bool(writer_args["treatWarningAsError"]),
            }
        new_writer_args = inchi_kwargs.copy()
        new_writer_args["delimiter"] = delimiter.name
        new_writer_args["include_id"] = include_id = bool(writer_args["include_id"])

        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_writer_args)
        return base_toolkit.MoleculeWriter(
            metadata,
            gen_write_inchi_structures(outfile, close, error_handler, location, delimiter,
                                       include_id, inchi_kwargs),
            location)
    

    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        inchi_kwargs = {
            "options": writer_args["options"],
            "logLevel": writer_args["logLevel"],
            "treatWarningAsError": bool(writer_args["treatWarningAsError"]),
            }
        new_writer_args = inchi_kwargs.copy()
        new_writer_args["delimiter"] = delimiter.name
        new_writer_args["include_id"] = include_id = bool(writer_args["include_id"])

        outfile, close, getvalue = io._open_output_string()
        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_inchi_structures(outfile, close, error_handler, location, delimiter,
                                       include_id, inchi_kwargs),
            getvalue,
            location)

    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        inchi_kwargs = {
            "options": writer_args["options"],
            "logLevel": writer_args["logLevel"],
            "treatWarningAsError": bool(writer_args["treatWarningAsError"]),
            }
        new_writer_args = inchi_kwargs.copy()
        new_writer_args["delimiter"] = delimiter.name
        new_writer_args["include_id"] = include_id = bool(writer_args["include_id"])

        outfile, close, getvalue = io._open_compressed_output_bytes(compression, EXTERNAL_NAME, level)
        
        metadata = base_toolkit.FormatMetadata(_or_bytes(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_inchi_structures(outfile, close, error_handler, location, delimiter,
                                       include_id, inchi_kwargs),
            getvalue,
            location)
                

    def create_string(self, mol, id, writer_args, error_handler):
        try:
            inchi = Chem.MolToInchi(mol,
                                    options = writer_args["options"],
                                    logLevel = writer_args["logLevel"],
                                    treatWarningAsError = writer_args["treatWarningAsError"],
                                    )
        except InchiReadWriteError as err:
            error_handler.error("RDKit cannot create the InChI string: %s" % (err.args[1],))
            return None
        except MolSanitizeException as err:
            error_handler.error("RDKit cannot create the InChI string: %s" % (err,))
            return None
        except ValueError as err:
            # It looks like the molecule is sanitized before passing
            # the molecule to InChI, and sanitization can raise a
            # ValueError. But so can passing in the wrong value.
            msg = str(err)
            if msg.startswith("Sanitization error"):
                error_handler.error("RDKit cannot create the InChI string: %s" % msg)
                return None
            else:
                raise
        except RuntimeError as err:
            error_handler.error("RDKit cannot create the InChI string: %s" % (err,))
            return None
        
        if not inchi:
            error_handler.error("RDKit cannot create the InChI string")
            return None
                    
        if writer_args["include_id"]:
            delimiter = base_toolkit.get_delimiter(self, writer_args)
            if id is None:
                try:
                    id = _GetProp(mol, u"_Name")
                except KeyError:
                    id = u""
            return delimiter.get_format_string() % (inchi, id)
        else:
            return inchi + u"\n"


class InChIKeyFormat(RDKitFormatConfig):
    register_name = record_format = "inchikey"
    is_available = has_inchi

    input = None

    output = InChIFormat.output

    def open_molecule_writer(self, destination, outfile, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        inchi_kwargs = {
            "options": writer_args["options"],
            "logLevel": writer_args["logLevel"],
            "treatWarningAsError": bool(writer_args["treatWarningAsError"]),
            }
        new_writer_args = inchi_kwargs.copy()
        new_writer_args["delimiter"] = delimiter.name
        new_writer_args["include_id"] = include_id = bool(writer_args["include_id"])

        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_writer_args)
        return base_toolkit.MoleculeWriter(
            metadata,
            gen_write_inchikey_structures(outfile, close, error_handler, location, delimiter,
                                          include_id, inchi_kwargs),
            location)

    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        inchi_kwargs = {
            "options": writer_args["options"],
            "logLevel": writer_args["logLevel"],
            "treatWarningAsError": bool(writer_args["treatWarningAsError"]),
            }
        new_writer_args = inchi_kwargs.copy()
        new_writer_args["delimiter"] = delimiter.name
        new_writer_args["include_id"] = include_id = bool(writer_args["include_id"])
        outfile, close, getvalue = io._open_output_string()
        
        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_inchikey_structures(outfile, close, error_handler, location, delimiter,
                                          include_id, inchi_kwargs),
            getvalue,
            location)
            
    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        inchi_kwargs = {
            "options": writer_args["options"],
            "logLevel": writer_args["logLevel"],
            "treatWarningAsError": bool(writer_args["treatWarningAsError"]),
            }
        new_writer_args = inchi_kwargs.copy()
        new_writer_args["delimiter"] = delimiter.name
        new_writer_args["include_id"] = include_id = bool(writer_args["include_id"])
        outfile, close, getvalue = io._open_compressed_output_bytes(compression, EXTERNAL_NAME, level)
        
        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_inchikey_structures(outfile, close, error_handler, location, delimiter,
                                          include_id, inchi_kwargs),
            getvalue,
            location)
            

    def create_string(self, mol, id, writer_args, error_handler):
        try:
            inchikey = _MolToInchiKey(mol,
                                      options = writer_args["options"],
                                      logLevel = writer_args["logLevel"],
                                      treatWarningAsError = writer_args["treatWarningAsError"],
                                      )
        except InchiReadWriteError as err:
            error_handler.error("RDKit cannot create the InChIKey string: %s" % (err.args[1],))
            return None
        except MolSanitizeException as err:
            error_handler.error("RDKit cannot create the InChIKey string: %s" % (err,))
            return None
        except ValueError as err:
            # It looks like the molecule is sanitized before passing
            # the molecule to InChI, and sanitization can raise a
            # ValueError. But so can passing in the wrong value.
            msg = str(err)
            if msg.startswith("Sanitization error"):
                error_handler.error("RDKit cannot create the InChIKey string: %s" % msg)
                return None
            else:
                raise
        if not inchikey:
            error_handler.error("RDKit cannot create the InChIKey string")
            return None
        if writer_args["include_id"]:
            delimiter = base_toolkit.get_delimiter(self, writer_args)
            format_string = delimiter.get_format_string()
            if id is None:
                try:
                    id = _GetProp(mol, u"_Name")
                except KeyError:
                    id = u""
            return format_string % (inchikey, id)
        else:
            return inchikey + u"\n"


def iter_read_sdf_structures(infile, close, sdf_kwargs, id_tag, error_handler, location,
                             encoding, encoding_errors):
    id_tag_bytes = _text_toolkit.get_id_tag_bytes(id_tag, encoding, encoding_errors)
    it = _iter_read_sdf_structures(infile, close, sdf_kwargs, id_tag_bytes, error_handler, location,
                                   encoding, encoding_errors)
    next(it)
    return it

def _iter_read_sdf_structures(infile, close, sdf_kwargs, id_tag_bytes, error_handler, location,
                              encoding, encoding_errors):
    # Register a way to get the current molecule
    mol = None
    def get_mol():
        return mol
    location.register(get_mol=get_mol)
    
    record_reader = _text_toolkit.iter_sdf_id_bytes_and_records(infile, close, id_tag_bytes, error_handler, location,
                                                                encoding, encoding_errors)
    try:
        yield "Ready for processing!"
        for (id_bytes, record) in record_reader:
            try:
                mol = MolFromSDBlock(record, **sdf_kwargs)
            except UnicodeDecodeError as err:
                error_handler.error("Cannot decode the record: %s" % (err,), location)
                continue
                
            if mol is None:
                # This was not a molecule?
                error_handler.error("Could not parse molecule block", location)
                continue

            if id_bytes is None:
                id = None
            else:
                try:
                    id = id_bytes.decode(encoding, encoding_errors)
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the identifier %r: %s" % (id_bytes, err), location)
                    continue
            
            yield id, mol
            mol = None  # reset in case the lower-level handler skips a record/raises an error
    finally:
        location.save(mol=None)
        mol = None


class SDFFormat(RDKitFormatConfig):
    register_name = record_format = "sdf"
    alternate_extensions = ("sd", "mdl")
    is_available = True

    class input(object):
        arg_info = (
            ("sanitize", base_toolkit.parse_boolean),
            ("removeHs", base_toolkit.parse_boolean),
            ("strictParsing", base_toolkit.parse_boolean),
            ("includeTags", base_toolkit.parse_boolean),
            )
            
        defaults = {
            "sanitize": True,
            "removeHs": True,
            "strictParsing": True,
            "includeTags": True,
        }

    class output(object):
        arg_info = (
            ("includeStereo", base_toolkit.parse_boolean),
            ("kekulize", base_toolkit.parse_boolean),
            ("v3k", base_toolkit.parse_boolean),
            )
        defaults = {
            "includeStereo": False,
            "kekulize": True,
            "v3k": False,
        }

    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler, location,
                               encoding, encoding_errors):
        infile, close = io.open_compressed_input(source, compression, EXTERNAL_NAME)

        new_reader_args = {
            "sanitize": bool(reader_args["sanitize"]),
            "removeHs": bool(reader_args["removeHs"]),
            "strictParsing": bool(reader_args["strictParsing"]),
            "includeTags": bool(reader_args["includeTags"]),
            }
                             
        iter_reader = iter_read_sdf_structures(infile, close, new_reader_args, id_tag, error_handler, location,
                                               encoding, encoding_errors)

        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_reader_args)
        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)

    def read_ids_and_molecules_from_string(self, content, compression,
                                            id_tag, reader_args, error_handler, location):
        assert not compression, compression
        infile = BytesIO(content)
        close = infile.close

        new_reader_args = {
            "sanitize": bool(reader_args["sanitize"]),
            "removeHs": bool(reader_args["removeHs"]),
            "strictParsing": bool(reader_args["strictParsing"]),
            "includeTags": bool(reader_args["includeTags"]),            
            }

        iter_reader = iter_read_sdf_structures(infile, close, new_reader_args, id_tag, error_handler, location,
                                               encoding="utf8", encoding_errors="strict")

        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format, new_reader_args)
        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        # Note: keep this synchronized with MolfileFormat
        id_tag_bytes = _text_toolkit.get_id_tag_bytes(id_tag, "utf8", "strict")

        if id_tag_bytes is None:
            def rdkit_sdf_parse_id_and_molecule(content):
                if isinstance(content, unicode_type):
                    content = content.encode("utf8")
                mol = MolFromSDBlock(content, **reader_args)
                if mol is None:
                    # This was not a molecule?
                    error_handler.error("RDKit cannot parse the sdf record")
                    return None, None
                try:
                    id = _GetProp(mol, u"_Name")
                except KeyError:
                    id = None
                return id, mol
        else:
            def rdkit_sdf_parse_id_and_molecule(content):
                if isinstance(content, unicode_type):
                    content = content.encode("utf8")
                mol = MolFromSDBlock(content, **reader_args)
                if mol is None:
                    # This was not a molecule?
                    error_handler.error("RDKit cannot parse the sdf record")
                    return None, None
                id = _text_toolkit.get_sdf_record_tag(content, id_tag_bytes)
                return id, mol

        return rdkit_sdf_parse_id_and_molecule

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        # Note: keep this synchronized with MolfileFormat
        if isinstance(content, unicode_type):
            if isinstance(id_tag, bytes_type):
                raise ValueError("id_tag must be a string, not bytes, when the content is a string")

        if not is_unicode_aware and isinstance(content, unicode_type):
            # Workaround for older versions of RDKit on Python 2.7
            content = content.encode("utf8")
            if id_tag is not None:
                id_tag = id_tag.encode("utf8")
        try:
            mol = MolFromSDBlock(content, **reader_args)
        except UnicodeDecodeError as err:
            error_handler.error("Could not decode the record: %s" % (err,))
            return None, None
        
        if mol is None:
            # This was not a molecule?
            error_handler.error("RDKit cannot parse the sdf record")
            return None, None

        if id_tag is None:
            try:
                id = _GetProp(mol, u"_Name")
            except KeyError:
                id = None
            
        else:
            id = _text_toolkit.get_sdf_record_tag(content, id_tag)
            
        return id, mol

    def open_molecule_writer(self, destination, outfile, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        new_writer_args = {
            "includeStereo": bool(writer_args["includeStereo"]),
            "kekulize": bool(writer_args["kekulize"]),
            "v3k": bool(writer_args["v3k"]),
            }
        
        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_writer_args)
        return base_toolkit.MoleculeWriter(
            metadata,
            gen_write_sdf_structures(outfile, close, error_handler, location, new_writer_args),
            location)
    
    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        outfile, close, getvalue = io._open_output_string()
        new_writer_args = {
            "includeStereo": bool(writer_args["includeStereo"]),
            "kekulize": bool(writer_args["kekulize"]),
            "v3k": bool(writer_args["v3k"]),
            }
        return base_toolkit.MoleculeStringWriter(
            base_toolkit.FormatMetadata(_or_string(location), self.record_format, new_writer_args),
            gen_write_sdf_structures(outfile, close, error_handler, location, new_writer_args),
            getvalue,
            location)
            
    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        outfile, close, getvalue = io._open_compressed_output_bytes(compression, EXTERNAL_NAME, level)
        new_writer_args = {
            "includeStereo": bool(writer_args["includeStereo"]),
            "kekulize": bool(writer_args["kekulize"]),
            "v3k": bool(writer_args["v3k"]),
            }
        return base_toolkit.MoleculeStringWriter(
            base_toolkit.FormatMetadata(_or_bytes(location), self.record_format, new_writer_args),
            gen_write_sdf_structures(outfile, close, error_handler, location, new_writer_args),
            getvalue,
            location)
            

    def create_string(self, mol, id, writer_args, error_handler):
        return u"".join(_mol_to_sd_block_lines(id, mol, writer_args.pop("v3k", False), writer_args))
    

class SmilesStringFormat(base_toolkit.NotIoFormatConfigMixin, RDKitFormatConfig):
    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        sanitize = reader_args["sanitize"]
        def rdkit_parser(content):
            # Follow the Daylight spec, which says that the SMILES ends with any whitespace
            terms = content.split(None, 1)
            if not terms:
                # In chemfp, the empty SMILES string returns an empty molecule.
                # (RDKit returns no molecule for this case.)
                return None, Chem.Mol()

            mol = MolFromSmiles(terms[0], sanitize)
            if mol is None:
                error_handler.error("RDKit cannot parse the SMILES string %s" % (_compat.myrepr(terms[0]),))

            # This is only supposed to have a SMILES. Ignore any possible id.
            return None, mol

        rdkit_parser.__name__ = "rdkit_" + self.register_name + "_parse_id_and_molecule"
        return rdkit_parser
    
    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        # Follow the Daylight spec, which says that the SMILES ends with any whitespace
        terms = content.split(None, 1)
        if not terms:
            # In chemfp, the empty SMILES string returns an empty molecule.
            # (RDKit returns no molecule for this case.)
            return None, Chem.Mol()

        mol = MolFromSmiles(terms[0], reader_args["sanitize"])
        if mol is None:
            error_handler.error("RDKit cannot parse the SMILES string %s" % (_compat.myrepr(terms[0]),))

        # This is only supposed to have a SMILES. Ignore any possible id.
        return None, mol


    def create_string(self, mol, id, writer_args, error_handler):
        if writer_args.pop("cxsmiles", False):
            ToSmiles = MolToCXSmiles
        else:
            ToSmiles = MolToSmiles
        return ToSmiles(mol, **writer_args)

    def create_bytes(self, mol, id, writer_args, error_handler):
        if writer_args.pop("cxsmiles", False):
            ToSmiles = MolToCXSmiles
        else:
            ToSmiles = MolToSmiles
        return ToSmiles(mol, **writer_args).encode("utf8")


class SmiStringFormat(SmilesStringFormat):
    register_name = record_format = "smistring"

    class input(object):
        arg_info = _smi_input_arg_info
        defaults = _smi_input_defaults.copy()

    class output(object):
        arg_info = _smi_output_arg_info
        defaults = _smi_output_defaults.copy()
        defaults.update({
            "isomericSmiles": True,   # Both isomeric ... (following OEChem-2.0 style)
            "canonical": True,         # .. and canonical (following RDKit style)
            })
        

class CanStringFormat(SmilesStringFormat):
    register_name = record_format = "canstring"
    
    input = SmiStringFormat.input

    class output(object):
        arg_info = SmiStringFormat.output.arg_info
        defaults = SmiStringFormat.output.defaults.copy()
        defaults.update({
            "isomericSmiles": False,
            "canonical": True,         # This is the canonical SMILES, but not isomeric
            })
            

class UsmStringFormat(SmilesStringFormat):
    register_name = record_format = "usmstring"
    
    input = SmiStringFormat.input

    class output(object):
        arg_info = SmiStringFormat.output.arg_info
        defaults = SmiStringFormat.output.defaults.copy()
        defaults.update({
            "isomericSmiles": False,    # not isomeric ...
            "canonical": False,         #   ... and not canonical
            })


class InChIStringFormat(base_toolkit.NotIoFormatConfigMixin, RDKitFormatConfig):
    register_name = record_format = "inchistring"
    is_available = has_inchi

    class input(object):
        arg_info = _inchi_input_arg_info
        defaults = _inchi_input_defaults.copy()
        
    class output(object):
        arg_info = _inchi_output_arg_info
        defaults = _inchi_output_defaults.copy()

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        inchi_kwargs = dict(
            sanitize = reader_args["sanitize"],
            removeHs = reader_args["removeHs"],
            logLevel = reader_args["logLevel"],
            treatWarningAsError = reader_args["treatWarningAsError"])
                                    
        def rdkit_inchistring_parse_id_and_molecule(content):
            # Follow the SMILES lead and split on the first whitespace. (Is that right?)
            terms = content.split()
            if not terms:
                error_handler.error("The empty string is not a valid InChI input")
                return None, None
            inchi = terms[0]
            mol = MolFromInchi(inchi, **inchi_kwargs)
            if mol is None:
                error_handler.error("RDKit cannot parse the InChI %r" % (inchi,))

            # This is only supposed to have an InChI string. Ignore any possible id.
            return None, mol

        return rdkit_inchistring_parse_id_and_molecule
    
    
    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        # Follow the SMILES lead and split on the first whitespace. (Is that right?)
        terms = content.split()
        if not terms:
            error_handler.error("The empty string is not a valid InChI input")
            return None, None
        inchi = terms[0]
        mol = MolFromInchi(inchi,
                           sanitize = reader_args["sanitize"],
                           removeHs = reader_args["removeHs"],
                           logLevel = reader_args["logLevel"],
                           treatWarningAsError = reader_args["treatWarningAsError"])
        if mol is None:
            error_handler.error("RDKit cannot parse the InChI %r" % (inchi,))

        # This is only supposed to have an InChI string. Ignore any possible id.
        return None, mol

    def create_string(self, mol, id, writer_args, error_handler):
        try:
            inchi = Chem.MolToInchi(mol, **writer_args)
        except InchiReadWriteError as err:
            error_handler.error("RDKit cannot create the InChI string: %s" % (err.args[1],))
            return None
        except ValueError as err:
            # It looks like the molecule is sanitized before passing
            # the molecule to InChI, and sanitization can raise a
            # ValueError. But so can passing in the wrong value.
            msg = str(err)
            if msg.startswith("Sanitization error"):
                error_handler.error("RDKit cannot create the InChI string: %s" % msg)
                return None
            else:
                raise
        except RuntimeError as err:
            error_handler.error("RDKit cannot create the InChI string: %s" % (err,))
            return None
        if not inchi:
            error_handler.error("RDKit cannot create the InChI string")
            return None
        if not isinstance(inchi, type(u"")):
            inchi = inchi.decode("utf8")
        return inchi

class InChIKeyStringFormat(base_toolkit.NotIoFormatConfigMixin, InChIKeyFormat):
    register_name = format_name = "inchikeystring"
    is_available = has_inchi    

    input = None
        
    class output(object):
        arg_info = _inchi_output_arg_info
        defaults = _inchi_output_defaults.copy()

    def create_string(self, mol, id, writer_args, error_handler):
        try:
            inchikey = _MolToInchiKey(mol, **writer_args)
        except InchiReadWriteError as err:
            error_handler.error("RDKit cannot create the InChIKey string: %s" % (err.args[1],))
            return None
        except ValueError as err:
            # It looks like the molecule is sanitized before passing
            # the molecule to InChI, and sanitization can raise a
            # ValueError. But so can passing in the wrong value.
            msg = str(err)
            if msg.startswith("Sanitization error"):
                error_handler.error("RDKit cannot create the InChI string: %s" % msg)
                return None
            else:
                raise
        except RuntimeError as err:
            error_handler.error("RDKit cannot create the InChIKey string: %s" % (err,))
            return None
        if not inchikey:
            error_handler.error("RDKit cannot create the InChIKey string")
            return None
        if not isinstance(inchikey, type(u"")):
            inchikey = inchikey.decode("utf8")
        return inchikey


class MolfileFormat(base_toolkit.NotIoFormatConfigMixin, RDKitFormatConfig):
    register_name = record_format = "molfile"
    alternate_extensions = ()
    is_available = True

    class input(object):
        arg_info = (
            ("sanitize", base_toolkit.parse_boolean),
            ("removeHs", base_toolkit.parse_boolean),
            ("strictParsing", base_toolkit.parse_boolean),
            )
            
        defaults = {
            "sanitize": True,
            "removeHs": True,
            "strictParsing": True,
        }

    output = SDFFormat.output

    
    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        # Note: keep this synchronized with SDFFormat
        def rdkit_molfile_parse_id_and_molecule(content):
            if isinstance(content, unicode_type):
                content = content.encode("utf8")
            mol = MolFromSDBlock(content, includeTags=False, **reader_args)
            if mol is None:
                # This was not a molecule?
                error_handler.error("RDKit cannot parse the molfile record")
                return None, None

            # molfiles don't have tag data, so leave it with the title line.
            try:
                id = _GetProp(mol, u"_Name")
            except KeyError:
                id = None

            return id, mol

        return rdkit_molfile_parse_id_and_molecule
        

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        # Note: keep this synchronized with SDFormat
        mol = MolFromSDBlock(content, includeTags=False, **reader_args)
        if mol is None:
            # This was not a molecule?
            error_handler.error("RDKit cannot parse the molfile record")
            return None, None

        # molfiles don't have tag data, so leave it with the title line.
        try:
            id = _GetProp(mol, u"_Name")
        except KeyError:
            id = None
            
        return id, mol
        
    def create_string(self, mol, id, writer_args, error_handler):
        if writer_args.pop("v3k", False):
            block = MolToV3KMolBlock(mol, **writer_args)
        else:
            block = Chem.MolToMolBlock(mol, **writer_args)
        if id is None:
            return block
        else:
            left, mid, right = block.partition("\n")
            return id + mid + right


class RDBinMolFormat(base_toolkit.NotIoFormatConfigMixin, RDKitFormatConfig):
    register_name = record_format = "rdbinmol"
    is_available = True

    class input(object):
        arg_info = ()
        defaults = {}
    
    class output(object):
        arg_info = ()
        defaults = {}

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        def rdkit_rdbinmol_parse_id_and_molecule(content):
            try:
                mol = Chem.Mol(content)
            except Exception:
                error_handler.error("RDKit cannot parse the rdbinmol record")
                return None, None
            try:
                id = _GetProp(mol, u"_Name")
            except KeyError:
                id = None

            return id, mol
        return rdkit_rdbinmol_parse_id_and_molecule

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        try:
            mol = Chem.Mol(content)
        except Exception:
            error_handler.error("RDKit cannot parse the rdbinmol record")
            return None, None

        try:
            id = _GetProp(mol, u"_Name")
        except KeyError:
            id = None

        return id, mol

    def create_string(self, mol, id, writer_args, error_handler):
        if id is None:
            return mol.ToBinary()
        
        # Special-case if the id is already the record's title
        old_id = _GetProp(mol, u"_Name")
        if old_id == id:
            return mol.ToBinary()
        
        # The only thread-safe solution is to make a copy and
        # set the title on the copy.
        mol = Chem.Mol(mol.ToBinary())
        _SetProp(mol, u"_Name", id)
        return mol.ToBinary()

    def create_bytes(self, mol, id, writer_args, error_handler):
        if id is None:
            return mol.ToBinary()
        
        # Special-case if the id is already the record's title
        old_id = _GetProp(mol, u"_Name")
        if old_id == id:
            return mol.ToBinary()
        
        # The only thread-safe solution is to make a copy and
        # set the title on the copy.
        mol = Chem.Mol(mol.ToBinary())
        _SetProp(mol, u"_Name", id)
        return mol.ToBinary()


################### SDF-related I/O


# RDKit doesn't have a way to write a molecule to an SD record. (?)
# Use MolToMolBlock to get the connection table, then manually
# add the tags myself.
def _mol_to_sd_block_lines(id, mol, is_v3k, writer_args):
    if is_v3k:
        block = MolToV3KMolBlock(mol, **writer_args)
    else:
        block = Chem.MolToMolBlock(mol, **writer_args)
    if not is_unicode_aware:
        block = block.decode("utf8")
    if id is None:
        lines = [block]
    else:
        left, mid, right = block.partition(u"\n")
        lines = [id+mid, right]
    if is_unicode_aware:
        for name in mol.GetPropNames():
            lines.append(u"> <%s>\n%s\n\n" % (name, mol.GetProp(name)))
    else:
        # Compatibility for older RDKit versions under Python 2.7
        for name in mol.GetPropNames():
            text = "> <%s>\n%s\n\n" % (name, mol.GetProp(name))
            text = unicode(text, "utf8")
            lines.append(text)
        
    lines.append(u"$$$$\n")
    return lines

# Output coroutine for SDF writer
def gen_write_sdf_structures(outfile, close, error_handler, location, sdf_kwargs):
    gen = _gen_write_sdf_structures(outfile, close, error_handler, location, dict(sdf_kwargs))
    next(gen)
    return gen

def _gen_write_sdf_structures(outfile, close, error_handler, location, sdf_kwargs):
    recno = output_recno = 0
    def get_recno():
        return recno
    def get_output_recno():
        return output_recno
    location.register(get_recno=get_recno,
                      get_output_recno=get_output_recno)
    is_v3k = sdf_kwargs.pop("v3k", False)
    try:
        yield_value = None
        while 1:
            command, args = yield yield_value
            yield_value = None
            try:
                if command == "write_molecules":
                    for mol in args:
                        recno += 1
                        lines = _mol_to_sd_block_lines(None, mol, is_v3k, sdf_kwargs)
                        if not outfile.writelines(lines, error_handler, location):
                            continue
                        output_recno += 1

                elif command == "write_ids_and_molecules":
                    for id, mol in args:
                        recno += 1
                        lines = _mol_to_sd_block_lines(id, mol, is_v3k, sdf_kwargs)
                        if not outfile.writelines(lines, error_handler, location):
                            continue
                        output_recno += 1

                elif command == "write_molecule":
                    id, mol = args
                    recno += 1
                    lines = _mol_to_sd_block_lines(id, mol, is_v3k, sdf_kwargs)
                    if not outfile.writelines(lines, error_handler, location):
                        continue
                    output_recno += 1

                else:
                    raise AssertionError("Unknown command: %r" % (command,))
            except Exception as err:
                yield_value = err, sys.exc_info()[2]
                
    finally:
        if close is not None:
            close()
        location.save(recno = get_recno(),
                      output_recno = get_output_recno())



#### Sequence and FASTA

#  0 Protein, L amino acids (default)
#  1 Protein, D amino acids
#  2 RNA, no cap
#  3 RNA, 5' cap
#  4 RNA, 3' cap
#  5 RNA, both caps
#  6 DNA, no cap
#  7 DNA, 5' cap
#  8 DNA, 3' cap
#  9 DNA, both caps

# Could have aliases, like:
# AA
# AA-L
# AA-D
# RNA-NoCap
# RNA-5pCap
# RNA-3pCap
# RNA-Caps
# DNA-NoCap
# DNA-5pCap
# DNA-3pCap
# DNA-Caps
        
def _parse_fasta_input_flavor(name, value):
    if value.isdigit() and len(value) == 1:
        v = int(value)
        if 0 <= 0 <= 9:
            return v
    raise ValueError("%s flavor must be between 0 and 9, inclusive, not %r" % (
        name, value))

def parse_sequence_input_flavor(value):
    return _parse_fasta_input_flavor("Sequence", value)

def parse_fasta_input_flavor(value):
    return _parse_fasta_input_flavor("FASTA", value)

class SequenceFormat(base_toolkit.NotIoFormatConfigMixin, RDKitFormatConfig):
    register_name = record_Format = "sequence"
    is_available = True

    class input(object):
        arg_info = (
            ("sanitize", base_toolkit.parse_boolean),
            ("flavor", parse_sequence_input_flavor),
            )
        defaults = {
            "sanitize": True,
            "flavor": 0,
            }
        
    class output(object):
        arg_info = ()
        defaults = {}

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        sanitize = reader_args["sanitize"]
        flavor = reader_args["flavor"]
        def rdkit_sequence_parse_id_and_molecule(content):
            mol = Chem.MolFromSequence(content, sanitize, flavor)
            if mol is None:
                if len(content) > 50:
                    error_handler.error("RDKit cannot parse the sequence starting %s" % (
                        _compat.myrepr(content[:45],)))
                else:
                    error_handler.error("RDKit cannot parse the sequence %s" % (
                        _compat.myrepr(content[:45],)))
                return None, None
            else:
                # sequences don't have an id
                return None, mol
        return rdkit_sequence_parse_id_and_molecule

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        mol = Chem.MolFromSequence(content, reader_args["sanitize"], reader_args["flavor"])
        if mol is None:
            error_handler.error("RDKit cannot parse the sequence record")
            return None, None
        else:
            # sequences don't have an id
            return None, mol

    def create_string(self, mol, id, writer_args, error_handler):
        seq = Chem.MolToSequence(mol)
        if not seq:
            # XXX Should I check and report if the atoms are missing a GetMonomerInfo()?
            error_handler.error("RDKit could not convert the molecule to a sequence")
            return None
        return seq


def _parse_fasta_record(record, sanitize, flavor, error_handler, location):
    # RDKit has a bug (repeated 2020-04-09 at https://github.com/rdkit/rdkit/issues/3080 )
    # where FASTA records with a title and invalid sequence cause a segmentation fault.
    #   
    # Instead, parse the record myself
    c = record[:1]
    if c == u">":
        title_line, _, seq = record.partition(u"\n")
        if title_line[-1:] == u"\r":
            title = title_line[1:-1]
        else:
            title = title_line[1:]
        mol = Chem.MolFromSequence(seq, sanitize, flavor)
        if mol is not None:
            mol.SetProp("_Name", title)
            return title, mol
        # error falls through
    elif c == b">":
        title_line, _, seq = record.partition(b"\n")
        if title_line[-1:] == b"\r":
            title = title_line[1:-1]
        else:
            title = title_line[1:]
        mol = Chem.MolFromSequence(seq, sanitize, flavor)
        if mol is not None:
            mol.SetProp("_Name", title)
            return title.decode("utf8"), mol
        # error falls through
    else:
        mol = Chem.MolFromSequence(record)
        if mol is not None:
            return None, mol
        # error falls through

    # Handle all errors
    error_handler.error("RDKit cannot parse the FASTA record", location)
    return None, None
        
            

    #### This is the old code
    ## mol = Chem.MolFromFASTA(record, sanitize, flavor)
    ## if mol is None:
    ##     if len(record) > 50:
    ##         error_handler.error("RDKit cannot parse the FASTA record starting %s" % (
    ##             _compat.myrepr(record[:45]),), location)
    ##     else:
    ##         error_handler.error("RDKit cannot parse the FASTA record %s" % (
    ##             _compat.myrepr(record),), location)
    ##     return None, None

    ## if record[:1] == b">":
    ##     return mol.GetProp("_Name"), mol
    ## elif record[:1] == u">":
    ##     return mol.GetProp("_Name"), mol
    ## else:
    ##     return None, mol
    
def iter_read_fasta_structures(
        infile, close, sanitize, flavor,
        error_handler, location, encoding, encoding_errors):
    it = _iter_read_fasta_structures(infile, close, sanitize, flavor,
                                       error_handler, location, encoding, encoding_errors)
    next(it)
    return it

def _iter_read_fasta_structures(
        infile, close, sanitize, flavor,
        error_handler, location, encoding, encoding_errors):
    recno = lineno = start_offset = end_offset = 0
    record = None
    mol = None
    def get_lineno():
        return lineno
    def get_recno():
        return recno
    def get_mol():
        return mol
    def get_offsets():
        return start_offset, end_offset
    def get_record():
        return record
    
    location.register(get_lineno = get_lineno,
                      get_recno = get_recno,
                      get_mol = get_mol,
                      get_offsets = get_offsets,
                      get_record = get_record,
        )
    lines = []
    offset = 0
    try:
        yield "ready"
        it = enumerate(infile, 1)
        lineno_line = next(it, None)
        if lineno_line is not None:
            lineno, line = lineno_line
            start_lineno = lineno
            offset = len(line)
            if line[:1] != b">":
                end_offset = offset
                error_handler.error("Not a FASTA file - first line must start with a '>'",
                                        location)
            else:
                lines = [line]
                recno = 1
                for lineno, line in it:
                    if line[:1] == b">":
                        record = b"".join(lines)
                        end_offset = offset
                        old_lineno, lineno = lineno, start_lineno
                        id, mol = _parse_fasta_record(record, sanitize, flavor, error_handler, location)
                        if mol is not None:
                            yield id, mol
                        mol = None
                        start_lineno = lineno = old_lineno
                        start_offset = end_offset
                        recno += 1
                        lines = []
                    lines.append(line)
                    offset += len(line)
                    
                if lines:
                    record = b"".join(lines)
                    old_lineno, lineno = lineno, start_lineno
                    yield _parse_fasta_record(record, sanitize, flavor, error_handler, location)
                    lineno = old_lineno
    
    finally:
        if close is not None:
            close()
        location.save(
            lineno=get_lineno(),
            recno=get_recno(),
            mol = None,
            offsets = None,
            record = None,
            )              

# Output coroutine for FASTA writer

def _create_fasta_record(id, mol, error_handler, location=None):
    record = Chem.MolToFASTA(mol)
    if not record:
        # XXX Should I check and report if the atoms are missing a GetMonomerInfo()?
        error_handler.error("RDKit could not convert the molecule to FASTA format",
                                location)
        return None
    if id is None:
        return record
    # It looks like RDKit always generates an id, which is good.
    assert record[:1] == ">", record
    # Remove the old id
    old_id, mid, seq = record.partition("\n")
    return ">" + id.replace("\n", "") + "\n" + seq


def gen_write_fasta_structures(outfile, close, error_handler, location):
    gen = _gen_write_fasta_structures(outfile, close, error_handler, location)
    next(gen)
    return gen

def _gen_write_fasta_structures(outfile, close, error_handler, location):
    recno = output_recno = 0
    def get_recno():
        return recno
    def get_output_recno():
        return output_recno
    location.register(get_recno=get_recno,
                      get_output_recno=get_output_recno)
    
    try:
        yield_value = None
        while 1:
            command, args = yield yield_value
            yield_value = None
            try:
                if command == "write_molecules":
                    for mol in args:
                        recno += 1
                        record = _create_fasta_record(None, mol, error_handler, location)
                        if record is not None:
                            if outfile.write_string(record, error_handler, location):
                                output_recno += 1

                elif command == "write_ids_and_molecules":
                    for id, mol in args:
                        recno += 1
                        record = _create_fasta_record(id, mol, error_handler, location)
                        if record is not None:
                            if outfile.write_string(record, error_handler, location):
                                output_recno += 1

                elif command == "write_molecule":
                    id, mol = args
                    recno += 1
                    record = _create_fasta_record(id, mol, error_handler, location)
                    if record is not None:
                        if outfile.write_string(record, error_handler, location):
                            output_recno += 1

                else:
                    raise AssertionError("Unknown command: %r" % (command,))
            except Exception as err:
                yield_value = err, sys.exc_info()[2]
                
    finally:
        if close is not None:
            close()
        location.save(recno = get_recno(),
                      output_recno = get_output_recno())

# Biopython has some test files with these extensions:
#   faa for amino acids
#   fna for nucleic acids
#   ffn for nucleic acids
# I haven't looked into if they are common.
        
class BaseFASTAFormat(RDKitFormatConfig):
    class input(object):
        arg_info = (
            ("sanitize", base_toolkit.parse_boolean),
            ("flavor", parse_fasta_input_flavor),
            )
        defaults = {
            "sanitize": True,
            "flavor": 0,
            }
        @staticmethod
        def with_defaults(sanitize=defaults["sanitize"], flavor=defaults["flavor"]):
            class input:
                arg_info = BaseFASTAFormat.input.arg_info
                defaults = {"sanitize": sanitize, "flavor": flavor}
            return input
        
    class output(object):
        arg_info = ()
        defaults = {}
        @staticmethod
        def with_defaults():
            class output:
                arg_info = BaseFASTAFormat.output.arg_info
                defaults = BaseFASTAFormat.output.defaults
            return output

    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler, location,
                               encoding, encoding_errors):
        infile, close = io.open_compressed_input(source, compression, EXTERNAL_NAME)

        sanitize = reader_args["sanitize"]
        flavor = reader_args["flavor"]
        new_reader_args = {
            "sanitize": sanitize,
            "flavor": flavor,
            }
                             
        iter_reader = iter_read_fasta_structures(infile, close, sanitize, flavor,
                                                    error_handler, location, encoding, encoding_errors)
        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_reader_args)
        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)

    def read_ids_and_molecules_from_string(self, content, compression,
                                            id_tag, reader_args, error_handler, location):
        infile = BytesIO(content)
        close = infile.close

        sanitize = reader_args["sanitize"]
        flavor = reader_args["flavor"]
        new_reader_args = {
            "sanitize": sanitize,
            "flavor": flavor,
            }

        iter_reader = iter_read_fasta_structures(infile, close, sanitize, flavor, error_handler, location,
                                               encoding="utf8", encoding_errors="strict")

        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format, new_reader_args)
        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        sanitize = reader_args["sanitize"]
        flavor = reader_args["flavor"]
        def rdkit_fasta_parse_id_and_molecule(record):
            return _parse_fasta_record(record, sanitize, flavor, error_handler, None)
        return rdkit_fasta_parse_id_and_molecule

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        return _parse_fasta_record(content, reader_args["sanitize"], reader_args["flavor"],
                                       error_handler, None)

    def open_molecule_writer(self, destination, outfile, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, {})
        
        return base_toolkit.MoleculeWriter(
            metadata,
            gen_write_fasta_structures(outfile, close, error_handler, location),
            location)
            

    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        outfile, close, getvalue = io._open_output_string()

        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format, {})
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_fasta_structures(outfile, close, error_handler, location),
            getvalue, location)
            

    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        outfile, close, getvalue = io._open_compressed_output_bytes(compression, EXTERNAL_NAME, level)
    
        metadata = base_toolkit.FormatMetadata(_or_bytes(location), self.record_format, {})
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_fasta_structures(outfile, close, error_handler, location),
            getvalue,
            location)
    
    def create_string(self, mol, id, writer_args, error_handler):
        return _create_fasta_record(id, mol, error_handler)

    
class FASTAFormat(BaseFASTAFormat):
    register_name = record_format = "fasta"
    is_available = True
    alternate_extensions = ("faa",) # XX any other extensions?

    input = BaseFASTAFormat.input.with_defaults()
    output = BaseFASTAFormat.output.with_defaults()

# The thought is to support ".fna" if people want a nucleic acid format
# But there's no demand for that, and I don't want to choose the right default.

## class FNAFormat(BaseFASTAFormat):
##     register_name = record_Format = "fna"
##     is_available = True

##     input = BaseFASTAFormat.input.with_defaults(flavor=???)
##     output = BaseFASTAFormat.output.with_defaults()


    
######## HELM
            
class HELMFormat(base_toolkit.NotIoFormatConfigMixin, RDKitFormatConfig):
    register_name = record_Foramt = "helm"
    is_available = True

    class input(object):
        arg_info = (
            ("sanitize", base_toolkit.parse_boolean),
            )
        defaults = {
            "sanitize": True,
            }
        
    class output(object):
        arg_info = ()
        defaults = {}

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        sanitize = reader_args["sanitize"]
        def rdkit_helm_parse_id_and_molecule(content):
            mol = Chem.MolFromHELM(content, sanitize)
            if mol is None:
                error_handler.error("RDKit cannot parse the HELM record")
                return None, None
            return None, mol
        
        return rdkit_helm_parse_id_and_molecule

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        mol = Chem.MolFromHELM(content, reader_args["sanitize"])
        if mol is None:
            error_handler.error("RDKit cannot parse the HELM record")
            return None, None
        return None, mol

    def create_string(self, mol, id, writer_args, error_handler):
        helm = Chem.MolToHELM(mol)
        if not helm:
            # XXX Should I check and report if the atoms are missing a GetMonomerInfo()?
            error_handler.error("RDKit could not convert the molecule to HELM format")
            return None

        return helm


##### Mol2

def iter_mol2_ids_and_mols(
        infile, close,
        sanitize, removeHs, cleanupSubstructures,
        error_handler, location, encoding, encoding_errors):
    it = _iter_mol2_ids_and_mols(
        infile, close,
        sanitize, removeHs, cleanupSubstructures,
        error_handler, location, encoding, encoding_errors)
    next(it)
    return it

def _iter_mol2_ids_and_mols(
        infile, close,
        sanitize, removeHs, cleanupSubstructures,
        error_handler, location, encoding, encoding_errors):
    mol = None
    recno = lineno = start_lineno = start_offset = end_offset = 0
    record = None
    def get_mol():
        return mol
    def get_lineno():
        return start_lineno
    def get_recno():
        return recno
    def get_offsets():
        return start_offset, end_offset
    def get_record():
        return record

    location.register(get_mol = get_mol,
                      get_lineno = get_lineno,
                      get_recno = get_recno,
                      get_offsets = get_offsets,
                      get_record = get_record,
        )
    lines = []
    found_first_molecule = False
    has_molecule = False
    title_line = None
    offset = 0
    try:
        yield "Ready for processing!"
        for lineno, line in enumerate(infile, 1):
            if line.rstrip() == b"@<TRIPOS>MOLECULE":
                has_molecule = True
                if found_first_molecule:
                    end_offset = offset
                    recno += 1
                    record = b"".join(lines)
                    mol = MolFromMol2Block(record, sanitize, removeHs, cleanupSubstructures)
                    if mol:
                        yield mol.GetProp("_Name"), mol
                    else:
                        if title_line is not None:
                            title_line = title_line.strip()
                        if not title_line:
                            error_handler.error("RDKit could not parse mol2 record", location)
                        else:
                            try:
                                title_line = title_line.decode(encoding, encoding_errors)
                            except UnicodeDecodeError:
                                pass
                            error_handler.error("RDKit could not parse mol2 record with title %r" % (title_line,), location)
                    start_offset = end_offset
                    start_lineno = lineno
                    lines = []
                else:
                    start_lineno = 1
                    found_first_molecule = True
                title_line = None
                
            elif title_line is None and has_molecule:
                # Get the line after "@<TRIPOS>MOLECULE"
                title_line = line
            
            offset += len(line)
            lines.append(line)

        if lines and has_molecule:
            end_offset = offset
            recno += 1
            record = b"".join(lines)
            mol = MolFromMol2Block(record, sanitize, removeHs, cleanupSubstructures)
            if mol:
                yield mol.GetProp("_Name"), mol
            else:
                if title_line is not None:
                    title_line = title_line.strip()
                if not title_line:
                    error_handler.error("RDKit could not parse mol2 record", location)
                else:
                    try:
                        title_line = title_line.decode(encoding, encoding_errors)
                    except UnicodeDecodeError:
                        pass
                    error_handler.error("RDKit could not parse mol2 record with title %r" % (title_line,), location)
        
        start_lineno = lineno
    finally:
        if close is not None:
            close()
        location.save(mol = None,
                      lineno = get_lineno(),
                      recno = get_recno(),
                      offsets = None,
                      record = None,
                      )
    
class Mol2Format(RDKitFormatConfig):
    register_name = record_format = "mol2"
    is_available = True

    ## Note: the RDKit documentation (2020) says:
    # "The parser expects the atom-typing scheme used by Corina. Atom
    # types from Tripos' dbtranslate are less supported. Other atom
    # typing schemes are unlikely to work."
    class input(object):
        arg_info = (
            ("sanitize", base_toolkit.parse_boolean),
            ("removeHs", base_toolkit.parse_boolean),
            ("cleanupSubstructures", base_toolkit.parse_boolean),
            )
        defaults = {
            "sanitize": True,
            "removeHs": True,
            "cleanupSubstructures": True,
            }

    output = None # RDKit doesn't support Mol2 output

    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler, location,
                                   encoding, encoding_errors):
        sanitize = bool(reader_args["sanitize"])
        removeHs = bool(reader_args["removeHs"])
        cleanupSubstructures = bool(reader_args["cleanupSubstructures"])

        infile, close = io.open_compressed_input(source, compression, EXTERNAL_NAME)
        iter_reader = iter_mol2_ids_and_mols(
            infile, close,
            sanitize, removeHs, cleanupSubstructures,
            error_handler, location, encoding, encoding_errors)
        metadata = base_toolkit.FormatMetadata(
            location.filename, self.record_format,
            {"sanitize": sanitize,
             "removeHs": removeHs,
              "cleanupSubstructures": cleanupSubstructures,
            })
        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)

    def read_ids_and_molecules_from_string(self, content, compression,
                                               id_tag, reader_args, error_handler, location):
        assert not compression, compression

        sanitize = bool(reader_args["sanitize"])
        removeHs = bool(reader_args["removeHs"])
        cleanupSubstructures = bool(reader_args["cleanupSubstructures"])

        infile = BytesIO(content)
        close = infile.close
        new_reader_args = {
            "sanitize": sanitize,
            "removeHs": removeHs,
            "cleanupSubstructures": cleanupSubstructures,
            }
        iter_reader = iter_mol2_ids_and_mols(
            infile, close,
            sanitize, removeHs, cleanupSubstructures,
            error_handler, location, encoding="utf8", encoding_errors="strict")
        metadata = base_toolkit.FormatMetadata(
            location.filename, self.record_format,
            new_reader_args)
        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        sanitize = bool(reader_args["sanitize"])
        removeHs = bool(reader_args["removeHs"])
        cleanupSubstructures = bool(reader_args["cleanupSubstructures"])

        def rdkit_mol2_parse_id_and_molecule(content):
            mol = MolFromMol2Block(content, sanitize, removeHs, cleanupSubstructures)
            if mol is None:
                error_handler.error("RDKit could not parse mol2 record")
            return mol.GetProp("_Name"), mol

        return rdkit_mol2_parse_id_and_molecule

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        mol = MolFromMol2Block(
            content,
            bool(reader_args["sanitize"]),
            bool(reader_args["removeHs"]),
            bool(reader_args["cleanupSubstructures"]),
            )
        if mol is None:
            error_handler.error("RDKit could not parse mol2 record")
            return None, None
        return mol.GetProp("_Name"), mol


##### PDB
    
def parse_pdb_input_flavor(value):
    if value.isdigit() and int(value) == 0:
        return 0
    raise ValueError("PDB input flavor must be 0")

# flavor & 1 : Write MODEL/ENDMDL lines around each record
# flavor & 2 : Don't write any CONECT records
# flavor & 4 : Write CONECT records in both directions
# flavor & 8 : Don't use multiple CONECTs to encode bond order
# flavor & 16 : Write MASTER record
# flavor & 32 : Write TER record

class Flavor(object):
    def __init__(self, register_name, canonical_options, flavors):
        self.register_name = register_name
        self.canonical_options = canonical_options
        self.flavors = flavors

    def parse_flavor(self, value_string):
        if not value_string:
            return self.flavors["Default"]
        
        if value_string.isdigit():
            return int(value_string)

        flavor = 0
        value_string = value_string.replace(",", "|") # Allow both ',' and '|' as separators
        try:
            for term in value_string.split("|"):
                if term[:1] == "-":
                    term_flavor = self.flavors[term[1:]]
                    flavor &= ~term_flavor
                else:
                    term_flavor = self.flavors[term]
                    flavor |= term_flavor
        except KeyError:
            available_flavors = ", ".join(self.canonical_options)
            raise ValueError(
                "RDKit %s format does not support the %r flavor option. Available flavors are: %s"
                % (self.register_name, term, available_flavors))

        return flavor

    def get_flavor(self, flavor):
        if isinstance(flavor, int):
            if not (0 <= flavor <= 2**31-1):
                raise ValueError("PDB output flavor value must be 0 <= flavor < 2**31")
            return flavor
        return self.parse_flavor(flavor)

_pdb_output_flavor = Flavor(
    "pdb output",
    canonical_options = (
        "MODEL", "NO_CONECT", "BOTH_CONECT", "NO_BOND_ORDER", "MASTER", "TER"),
    flavors = {
        "MODEL": 1,
        "NO_CONECT": 2,
        "BOTH_CONECT": 4,
        "NO_BOND_ORDER": 8,
        "MASTER": 16,
        "TER": 32,
        "Default": 0,
        "DEFAULT": 0,
        }
        )
    
def parse_pdb_output_flavor(value):
    return _pdb_output_flavor.parse_flavor(value)


def iter_pdb_ids_and_mols(
        infile, close,
        sanitize, removeHs, flavor, proximityBonding,
        error_handler, location, encoding, encoding_errors):
    it = _iter_pdb_ids_and_mols(
        infile, close,
        sanitize, removeHs, flavor, proximityBonding,
        error_handler, location, encoding, encoding_errors)
    next(it)
    return it

def _iter_pdb_ids_and_mols(
        infile, close,
        sanitize, removeHs, flavor, proximityBonding,
        error_handler, location, encoding, encoding_errors):
    mol = None
    recno = lineno = start_offset = end_offset = 0

    def get_mol():
        return mol
    def get_lineno():
        return lineno
    def get_recno():
        return recno
    def get_offsets():
        return start_offset, end_offset
    def get_record():
        return record
    
    location.register(get_mol = get_mol,
                      get_lineno = get_lineno,
                      get_recno = get_recno,
                      get_offsets = get_offsets,
                      get_record = get_record,
        )
    try:
        lines = []
        offset = 0
        start_lineno = 1
        yield "blah"
        for lineno, line in enumerate(infile, 1):
            offset += len(line)
            lines.append(line)

            if ((line[:3] == b"END") and (line[:6].strip() == b"END")):
                end_offset = offset
                record = b"".join(lines)
                recno += 1
                old_lineno, lineno = lineno, start_lineno
                mol = Chem.MolFromPDBBlock(record, sanitize, removeHs, flavor, proximityBonding)
                if mol is not None:
                    try:
                        yield mol.GetProp("_Name"), mol
                    except KeyError:
                        yield None, mol
                else:
                    error_handler.error("RDKit could not parse pdb record", location)
                lineno = old_lineno
                start_lineno = lineno + 1
                lines = []
                start_offset = end_offset
                
        if lines:
            end_offset = offset
            record = b"".join(lines)
            old_lineno, lineno = lineno, start_lineno
            mol = Chem.MolFromPDBBlock(record, sanitize, removeHs, flavor, proximityBonding)
            recno += 1
            if mol is not None:
                try:
                    yield mol.GetProp("_Name"), mol
                except KeyError:
                    yield None, mol
            else:
                error_handler.error("RDKit could not parse pdb record", location)
            lineno = old_lineno
        
    finally:
        if close is not None:
            close()
        location.save(mol = None,
                      lineno = get_lineno(),
                      recno = get_recno(),
                      offsets = None,
                      record = None,
                      )
        
def _get_pdb_record(id, mol, flavor):
    if id is None:
        return Chem.MolToPDBBlock(mol, flavor=flavor)

    try:
        old_id = mol.GetProp("_Name")
    except KeyError:
        # Has no id. Make a copy
        mol = Chem.Mol(mol)
        mol.SetProp("_Name", id)
        return Chem.MolToPDBBlock(mol, flavor=flavor)

    if old_id == id:
        # No need to make a copy
        return Chem.MolToPDBBlock(mol, flavor=flavor)

    # Replace the id via a copy
    mol = Chem.Mol(mol)
    mol.SetProp("_Name", id)
    return Chem.MolToPDBBlock(mol, flavor=flavor)

def gen_write_pdb_structures(outfile, close, error_handler, location, flavor):
    gen = _gen_write_pdb_structures(outfile, close, error_handler, location, flavor)
    next(gen)
    return gen

def _gen_write_pdb_structures(outfile, close, error_handler, location, flavor):
    recno = output_recno = 0
    def get_recno():
        return recno
    def get_output_recno():
        return output_recno
    location.register(get_recno=get_recno,
                      get_output_recno=get_output_recno)
    
    try:
        yield_value = None
        while 1:
            command, args = yield yield_value
            yield_value = None
            try:
                if command == "write_molecules":
                    for mol in args:
                        recno += 1
                        record = Chem.MolToPDBBlock(mol, flavor=flavor)
                        outfile.write_string(record, error_handler, location)
                        output_recno += 1
                        continue

                elif command == "write_ids_and_molecules":
                    for id, mol in args:
                        recno += 1
                        record = _get_pdb_record(id, mol, flavor)

                        outfile.write_string(record, error_handler, location)
                        output_recno += 1
                        continue

                elif command == "write_molecule":
                    id, mol = args
                    recno += 1
                    record = _get_pdb_record(id, mol, flavor)
                    outfile.write_string(record, error_handler, location)
                    output_recno += 1
                    continue

                else:
                    raise AssertionError("Unknown command: %r" % (command,))
            except Exception as err:
                yield_value = err, sys.exc_info()[2]
                
    finally:
        if close is not None:
            close()
        location.save(recno = get_recno(),
                      output_recno = get_output_recno())
        

class PDBFormat(RDKitFormatConfig):
    register_name = record_format = "pdb"
    # Older versions of RDKit's PDB reader don't support disabling proximityBonding.
    # I don't want to support versions which are that old.
    is_available = ("proximityBonding" in Chem.MolFromPDBBlock.__doc__)
    alternate_extensions = ("ent",)

    class input(object):
        arg_info = (
            ("sanitize", base_toolkit.parse_boolean),
            ("removeHs", base_toolkit.parse_boolean),
            ("flavor", parse_pdb_input_flavor),
            ("proximityBonding", base_toolkit.parse_boolean),
            )
        defaults = {
            "sanitize": True,
            "removeHs": True,
            "flavor": 0,
            "proximityBonding": True,
            }
            
    class output(object):
        arg_info = (
            ("flavor", parse_pdb_output_flavor),
            )
        defaults = {
            "flavor": 0,
            }

    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler, location,
                                   encoding, encoding_errors):
        sanitize = reader_args["sanitize"]
        removeHs = reader_args["removeHs"]
        flavor = reader_args["flavor"]
        proximityBonding = reader_args["proximityBonding"]

        infile, close = io.open_compressed_input(source, compression, EXTERNAL_NAME)

        iter_reader = iter_pdb_ids_and_mols(
            infile, close,
            sanitize, removeHs, flavor, proximityBonding,
            error_handler, location, encoding, encoding_errors)
        metadata = base_toolkit.FormatMetadata(
            location.filename, self.record_format,
            {"sanitize": sanitize,
             "removeHs": removeHs,
             "flavor": flavor,
             "proximityBonding": proximityBonding,
            })
        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)
            
    def read_ids_and_molecules_from_string(self, content, compression,
                                               id_tag, reader_args, error_handler, location):
        assert not compression, compression

        sanitize = reader_args["sanitize"]
        removeHs = reader_args["removeHs"]
        flavor = reader_args["flavor"]
        proximityBonding = reader_args["proximityBonding"]

        infile = BytesIO(content)
        close = infile.close
        new_reader_args = {
            "sanitize": sanitize,
            "removeHs": removeHs,
            "flavor": flavor,
            "proximityBonding": proximityBonding,
            }
        iter_reader = iter_pdb_ids_and_mols(
            infile, close,
            sanitize, removeHs, flavor, proximityBonding,
            error_handler, location, encoding="utf8", encoding_errors="strict")
        metadata = base_toolkit.FormatMetadata(
            location.filename, self.record_format,
            new_reader_args)
        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)
    
    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        sanitize = reader_args["sanitize"]
        removeHs = reader_args["removeHs"]
        flavor = reader_args["flavor"]
        proximityBonding = reader_args["proximityBonding"]

        def rdkit_pdb_parse_id_and_molecule(content):
            mol = Chem.MolFromPDBBlock(content, sanitize, removeHs, flavor, proximityBonding)
            if mol is None:
                error_handler.error("RDKit could not parse pdb record")
                return None, mol
            try:
                return mol.GetProp("_Name"), mol
            except KeyError:
                return None, mol

        return rdkit_pdb_parse_id_and_molecule

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        sanitize = reader_args["sanitize"]
        removeHs = reader_args["removeHs"]
        flavor = reader_args["flavor"]
        proximityBonding = reader_args["proximityBonding"]

        mol = Chem.MolFromPDBBlock(content, sanitize, removeHs, flavor, proximityBonding)
        if mol is None:
            error_handler.error("RDKit could not parse pdb record")
            return None, mol
        try:
            return mol.GetProp("_Name"), mol
        except KeyError:
            return None, mol


    def open_molecule_writer(self, destination, outfile, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        flavor = _pdb_output_flavor.get_flavor(writer_args["flavor"])
        
        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, {"flavor": flavor})
        
        return base_toolkit.MoleculeWriter(
            metadata,
            gen_write_pdb_structures(outfile, close, error_handler, location, flavor),
            location)
            

    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        flavor = _pdb_output_flavor.get_flavor(writer_args["flavor"])
        outfile, close, getvalue = io._open_output_string()

        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format,
                                                   {"flavor": flavor})
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_pdb_structures(outfile, close, error_handler, location, flavor),
            getvalue, location)

    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        flavor = _pdb_output_flavor.get_flavor(writer_args["flavor"])
        outfile, close, getvalue = io._open_compressed_output_bytes(compression, EXTERNAL_NAME, level)
    
        metadata = base_toolkit.FormatMetadata(_or_bytes(location), self.record_format,
                                                   {"flavor": flavor})
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_pdb_structures(outfile, close, error_handler, location, flavor),
            getvalue,
            location)
            

    def create_string(self, mol, id, writer_args, error_handler):
        flavor = _pdb_output_flavor.get_flavor(writer_args["flavor"])
        return _get_pdb_record(id, mol, flavor)
                


##### XYZ

def gen_write_xyz_structures(outfile, close, error_handler, location):
    it = _gen_write_xyz_structures(outfile, close, error_handler, location)
    next(it)
    return it

def _gen_write_xyz_structures(outfile, close, error_handler, location):
    recno = output_recno = 0
    def get_recno():
        return recno
    def get_output_recno():
        return output_recno

    location.register(get_recno=get_recno,
                      get_output_recno=get_output_recno)
    try:
        yield_value = None
        while 1:
            command, args = yield yield_value
            yield_value = None
            try:
                if command == "write_molecules":
                    for mol in args:
                        recno += 1
                        record = _create_xyz_record(None, mol, error_handler, location)
                        if record is not None:
                            output_recno += 1
                            outfile.write_string(record, error_handler, location)

                elif command == "write_ids_and_molecules":
                    for id, mol in args:
                        recno += 1
                        record = _create_xyz_record(id, mol, error_handler, location)
                        if record is not None:
                            output_recno += 1
                            outfile.write_string(record, error_handler, location)
                            
                elif command == "write_molecule":
                    id, mol = args
                    recno += 1
                    record = _create_xyz_record(id, mol, error_handler, location)
                    if record is not None:
                        output_recno += 1
                        outfile.write_string(record, error_handler, location)
            except Exception as err:
                yield_value = err, sys.exc_info()[2]
                        
    finally:
        if close is not None:
            close()
        location.save(output_recno = get_output_recno(),
                      recno = get_recno())
        
    
def _create_xyz_record(id, mol, error_handler, location=None):
    if mol.GetNumAtoms() == 0:
        # I'll handle the no-atom case. Because, why not?
        if id is None:
            try:
                id = mol.GetProp("_Name")
            except KeyError:
                id = ""
        else:
            id = id.replace("\n", "")
        return "0\n" + id + "\n"
    
    if mol.GetNumConformers() == 0:
        error_handler.error(
            "Cannot write molecules with no conformers to XYZ block",
            location)
        return None
    
    record = Chem.MolToXYZBlock(mol)
    if not record:
        error_handler.error("Cannot convert molecule to XYZ block", location)
        return None
    
    if id is None:
        return record
    
    # Stick in the id
    lines = record.splitlines(True)
    if len(lines) > 2:
        lines[1] = id.replace("\n", "") + "\n"
        return "".join(lines)

    raise AssertionError("Unexpected result from MolToXYZBlock(): %r" % (record,))


class XYZFormat(RDKitFormatConfig):
    register_name = record_format = "xyz"
    is_available = hasattr(Chem, "MolToXYZBlock")

    input = None

    class output(object):
        arg_info = ()
        defaults = {}

    def open_molecule_writer(self, destination, outfile, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, {})
        return base_toolkit.MoleculeWriter(
            metadata,
            gen_write_xyz_structures(outfile, close, error_handler, location),
            location)

    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        outfile, close, getvalue = io._open_output_string()
        
        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format, {})
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_xyz_structures(outfile, close, error_handler, location),
            getvalue,
            location)
            
    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        outfile, close, getvalue = io._open_compressed_output_bytes(compression, EXTERNAL_NAME, level)
        
        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format, {})
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_xyz_structures(outfile, close, error_handler, location),
            getvalue,
            location)

    def create_string(self, mol, id, writer_args, error_handler):
        return _create_xyz_record(id, mol, error_handler)

##### MAE (Maestro) format

# Only supports input

def iter_mae_ids_and_molecules(
        infile, close,
        sanitize, removeHs,
        id_tag,
        error_handler, location, encoding="utf8", encoding_errors="strict"):
    it = _iter_mae_ids_and_molecules(infile, close, sanitize, removeHs,
                                         id_tag, error_handler, location)
    next(it)
    return it

def _iter_mae_ids_and_molecules(infile, close, sanitize, removeHs, id_tag,
                                    error_handler, location):
    suppl = None
    recno = 0
    def get_recno():
        return recno
    if id_tag is None:
        id_tag = "_Name"
    
    location.register(get_recno=get_recno)
    try:
        yield "ready!"
        error_msg = None
        try:
            suppl = Chem.MaeMolSupplier(infile, sanitize, removeHs)
        except RuntimeError as err:
            error_msg = str(err).rstrip("\n.")
        if error_msg is not None:
            recno += 1
            error_handler.error(error_msg, location)
            # Don't do anything. Can't parse the version
        else:
            try:
                for mol in suppl:
                    recno += 1
                    try:
                        id = mol.GetProp(id_tag)
                    except KeyError:
                        id = None
                    yield id, mol
                    
            except RuntimeError as err:
                # Fix the record number before reporting the error
                recno += 1
                error_msg = str(err).rstrip("\n.")
            if error_msg is not None:
                error_handler.error(error_msg, location)
                # can't recover
    finally:
        location.save(recno=get_recno())
        if close is not None:
            close()

    
class MAEFormat(RDKitFormatConfig):
    register_name = record_format = "mae"
    alternate_extensions = ("maegz",)  # In principle this allows maegz.gz. Oh well.
    is_available = hasattr(Chem, "MaeMolSupplier")

    class input(object):
        arg_info = (
            ("sanitize", base_toolkit.parse_boolean),
            ("removeHs", base_toolkit.parse_boolean),
        )
        defaults = {
            "sanitize": True,
            "removeHs": True,
        }

    output = None

    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler,
                                   location, encoding, encoding_errors):
        sanitize = bool(reader_args["sanitize"])
        removeHs = bool(reader_args["removeHs"])

        # If source is a filename and it isn't compressed then I could let
        # MaeMolSupplier() parse the file directly.
        # That seems tricky. Instead, I'll handle the I/O in Python

        if not compression:
            # Figure out if I need to interpret this as gzip
            if isinstance(source, type(u"")) and source.lower().endswith(".maegz"):
                compression = "gz"
            elif isinstance(source, type(b"")) and source.lower().endswith(b".maegz"):
                compression = "gz"
        
        infile, close = io.open_compressed_input(source, compression, EXTERNAL_NAME)

        iter_reader = iter_mae_ids_and_molecules(infile, close, sanitize, removeHs, id_tag,
                                                         error_handler, location, encoding, encoding_errors)
        metadata = base_toolkit.FormatMetadata(location.filename, self.record_format,
                                               {"sanitize": sanitize,
                                                "removeHs": removeHs})
        
        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)

    def read_ids_and_molecules_from_string(self, content, compression,
                                               id_tag, reader_args, error_handler, location):
        assert not compression, compression

        sanitize = bool(reader_args["sanitize"])
        removeHs = bool(reader_args["removeHs"])

        infile = BytesIO(content)
        close = infile.close
        new_reader_args = {
            "sanitize": sanitize,
            "removeHs": removeHs,
            }
        iter_reader = iter_mae_ids_and_molecules(
            infile, close,
            sanitize, removeHs, id_tag,
            error_handler, location, encoding="utf8", encoding_errors="strict")
        metadata = base_toolkit.FormatMetadata(
            location.filename, self.record_format,
            new_reader_args)
        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        sanitize = bool(reader_args["sanitize"])
        removeHs = bool(reader_args["removeHs"])
        if id_tag is None:
            id_tag = "_Name"

        def rdkit_mae_parse_id_and_molecule(content):
            if isinstance(content, type(u"")):
                bio = BytesIO(content.encode("utf8"))
            else:
                bio = BytesIO(content)
            try:
                for mol in Chem.MaeMolSupplier(bio, sanitize, removeHs):
                    try:
                        id = mol.GetProp(id_tag)
                    except KeyError:
                        id = None
                    return id, mol
            except RuntimeError as err:
                errmsg = "RDKit could not parse mae record: %s" % (err,)
            else:
                errmsg = "RDKit could not find an mae record"
                
            error_handler.error(errmsg)
            return None, None

        return rdkit_mae_parse_id_and_molecule

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        sanitize = bool(reader_args["sanitize"])
        removeHs = bool(reader_args["removeHs"])
        if id_tag is None:
            id_tag = "_Name"

        if isinstance(content, type(u"")):
            bio = BytesIO(content.encode("utf8"))
        else:
            bio = BytesIO(content)

        try:
            for mol in Chem.MaeMolSupplier(bio, sanitize, removeHs):
                try:
                    id = mol.GetProp(id_tag)
                except KeyError:
                    id = None
                return id, mol
        except RuntimeError as err:
            errmsg = "RDKit could not parse mae record: %s" % (err,)
        else:
            errmsg = "RDKit could not find an mae record"
        error_handler.error(errmsg)
        return None, None
        


#### RDKit also supports these formats
#### Should I support them? If so, how?
## MolFromRDKitSVG
## MolFromTPLBlock  # What is this format?
## MolFromTPLFile
## MolToTPLBlock
## MolToTPLFile
## TDTMolSupplier  # Who uses Daylight TDTs these days?
## TDTWriter

## Register the formats
    
format_registry = base_toolkit.FormatRegistry(EXTERNAL_NAME, "rdkit")

format_registry.register_formats(
    SmiFormat(),
    CanFormat(),
    UsmFormat(),
    SDFFormat(),
    SmiStringFormat(),
    CanStringFormat(),
    UsmStringFormat(),
    MolfileFormat(),
    RDBinMolFormat(),
    FASTAFormat(),
    SequenceFormat(),
    HELMFormat(),
    Mol2Format(),
    PDBFormat(),
    XYZFormat(),
    MAEFormat(),
    
    InChIFormat(),
    #InChIRecordFormat()  # 'inchi format + identifier'
    InChIKeyFormat(),
    InChIStringFormat(),
    InChIKeyStringFormat(),
    )
        
################ The toolkit API



class RDKitToolkit(base_toolkit.BaseToolkit):
    def _open_molecule_writer_impl(self, format_config, destination, compression,
                                   writer_args, error_handler, location,
                                   encoding, encoding_errors, level):
        # Open the file and pass to the format
        exc = None
        try:
            outfile, close = io.open_compressed_output_wrapped(destination, compression, EXTERNAL_NAME,
                                                               encoding, encoding_errors, level)
        except ValueError as err:
            exc = ValueError("RDKit does not support compression type %r: %s" % (compression, err))
        if exc is not None:
            raise exc

        return format_config.open_molecule_writer(destination, outfile, close, writer_args,
                                                  error_handler, location, encoding, encoding_errors)
    
    def _read_ids_and_molecules_from_string_impl(self,
                                                 format_config, content, compression, id_tag,
                                                 reader_args, error_handler, location):
        if isinstance(content, unicode_type):
            if compression:
                raise ValueError("Compression does not work when the content is a Unicode string. "
                                 "Use a byte string instead.")
            if id_tag is not None and not isinstance(id_tag, unicode_type):
                raise ValueError("id_tag must be a string, not bytes, when the content is a string")

            content = content.encode("utf8")
        else:
            # Special case support for gzip and zstandard support via the file-based reader
            if compression == "gz":
                infile = BytesIO(content)
                return format_config.read_ids_and_molecules(infile, compression, id_tag, reader_args,
                                                            error_handler, location, "utf8", "strict")
        
            if compression == "zst":
                infile = BytesIO(content)
                return format_config.read_ids_and_molecules(infile, compression, id_tag, reader_args,
                                                            error_handler, location, "utf8", "strict")

            elif compression:
                raise ValueError("RDKit does not support compression type %r" % (compression,))

        return format_config.read_ids_and_molecules_from_string(content, compression, id_tag, reader_args,
                                                                error_handler, location)

    def _check_read_encoding(self, format, encoding, encoding_errors):
        if format.name == "sdf":
            base_toolkit.only_utf8(format, encoding, encoding_errors)

    def _check_write_encoding(self, format, encoding, encoding_errors):
        if format.name == "sdf":
            base_toolkit.only_utf8(format, encoding, encoding_errors)


toolkit = RDKitToolkit(format_registry)
